# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from __future__ import annotations

import math
from typing import Any, Callable, Dict, List, Optional, Tuple, Union, cast

import numpy as np
import structlog
from structlog.stdlib import BoundLogger

from dioptra.sdk.object_detection.architectures.tensorflow_layers.backbones import (  # noqa: B950
    EfficientNetBackbone,
)
from dioptra.sdk.object_detection.bounding_boxes.postprocessing import (
    BoundingBoxesYOLOV1PostProcessing,
)

from .tensorflow_layers import (
    MobileNetV2Backbone,
    SimpleYOLOV1Detector,
    TwoHeadedYOLOV1Detector,
    YOLOV1Detector,
)

LOGGER: BoundLogger = structlog.stdlib.get_logger()


try:
    import tensorflow as tf
    from tensorflow import GradientTape, Tensor
    from tensorflow.keras import Model
    from tensorflow.keras.layers import Layer

except ImportError:  # pragma: nocover
    LOGGER.warn(
        "Unable to import one or more optional packages, functionality may be reduced",
        package="tensorflow",
    )


try:
    from art.defences.preprocessor.preprocessor import Preprocessor
    from art.estimators.object_detection.object_detector import (
        ObjectDetector as ARTObjectDetector,
    )
    from art.estimators.object_detection.utils import convert_pt_to_tf, convert_tf_to_pt
    from art.utils import CLIP_VALUES_TYPE

except ImportError:  # pragma: nocover
    LOGGER.warn(
        "Unable to import one or more optional packages, functionality may be reduced",
        package="art",
    )


class YOLOV1ObjectDetector(Model):
    def __init__(
        self,
        input_shape: tuple[int, int, int],
        n_bounding_boxes: int,
        n_classes: int,
        backbone: str = "efficientnetb4",
        detector: str = "two_headed",
        name: str = "yolo_v1_object_detector",
    ) -> None:
        super().__init__(name=name)
        self.backbone = self._set_backbone(backbone=backbone, input_shape=input_shape)
        self.detector = self._set_detector(
            detector=detector,
            grid_shape=self.backbone.output_grid_shape,
            n_bounding_boxes=n_bounding_boxes,
            n_classes=n_classes,
        )
        self.loss_tracker = tf.keras.metrics.Mean(name="loss")
        self.val_loss_tracker = tf.keras.metrics.Mean(name="loss")
        self._image_input_shape = input_shape

    @property
    def image_input_shape(self) -> tuple[int, int, int]:
        return self._image_input_shape

    @property
    def output_grid_shape(self):
        return self.backbone.output_grid_shape

    @property
    def metrics(self):
        return [self.loss_tracker, self.val_loss_tracker]

    def call(self, inputs, training=None):
        x = self.backbone(inputs)

        return self.detector(x, training=training)

    def train_step(self, data):
        x, y = data

        with GradientTape() as tape:
            y_pred = self(x, training=True)
            loss = self.loss(y, y_pred)

        trainable_vars = self.trainable_variables
        gradients = tape.gradient(loss, trainable_vars)

        self.optimizer.apply_gradients(zip(gradients, trainable_vars))
        self.loss_tracker.update_state(loss)

        return {"loss": self.loss_tracker.result()}

    def test_step(self, data):
        x, y = data

        y_pred = self(x, training=False)
        loss = self.loss(y, y_pred)

        self.val_loss_tracker.update_state(loss)

        return {"loss": self.val_loss_tracker.result()}

    @staticmethod
    def _set_backbone(
        backbone: str, input_shape: Optional[Tuple[int, int, int]]
    ) -> Model:
        backbone_registry: dict[str, Callable[[], EfficientNetBackbone]] = dict(
            efficientnetb0=lambda: EfficientNetBackbone(
                flavor="b0", input_shape=input_shape
            ),
            efficientnetb1=lambda: EfficientNetBackbone(
                flavor="b1", input_shape=input_shape
            ),
            efficientnetb2=lambda: EfficientNetBackbone(
                flavor="b2", input_shape=input_shape
            ),
            efficientnetb3=lambda: EfficientNetBackbone(
                flavor="b3", input_shape=input_shape
            ),
            efficientnetb4=lambda: EfficientNetBackbone(
                flavor="b4", input_shape=input_shape
            ),
            efficientnetb5=lambda: EfficientNetBackbone(
                flavor="b5", input_shape=input_shape
            ),
            efficientnetb6=lambda: EfficientNetBackbone(
                flavor="b6", input_shape=input_shape
            ),
            efficientnetb7=lambda: EfficientNetBackbone(
                flavor="b7", input_shape=input_shape
            ),
            mobilenetv2=lambda: MobileNetV2Backbone(input_shape=input_shape),
        )

        return backbone_registry[backbone.strip().lower()]()

    @staticmethod
    def _set_detector(
        detector: str,
        grid_shape: Tuple[int, int],
        n_bounding_boxes: int,
        n_classes: int,
    ) -> Layer:
        detector_registry: dict[str, Callable[[], Layer]] = dict(
            original=lambda: YOLOV1Detector(
                grid_shape=grid_shape,
                n_bounding_boxes=n_bounding_boxes,
                n_classes=n_classes,
            ),
            shallow=lambda: SimpleYOLOV1Detector(
                grid_shape=grid_shape,
                n_bounding_boxes=n_bounding_boxes,
                n_classes=n_classes,
            ),
            two_headed=lambda: TwoHeadedYOLOV1Detector(
                grid_shape=grid_shape,
                n_bounding_boxes=n_bounding_boxes,
                n_classes=n_classes,
            ),
        )

        return detector_registry[detector.strip().lower()]()


class ARTYOLOV1ObjectDetector(ARTObjectDetector):
    def __init__(
        self,
        model: YOLOV1ObjectDetector,
        n_classes: int,
        bounding_boxes_postprocessing: BoundingBoxesYOLOV1PostProcessing,
        clip_values: Optional[CLIP_VALUES_TYPE] = (0, 255),
        preprocessing_defences: Union[Preprocessor, List[Preprocessor], None] = None,
        **kwargs,
    ) -> None:
        self._bounding_boxes_postprocessing = bounding_boxes_postprocessing
        self._channels_first = False
        self._n_classes = n_classes
        super().__init__(
            model=model,
            clip_values=clip_values,
            preprocessing_defences=preprocessing_defences,
            **kwargs,
        )

    @property
    def channels_first(self) -> bool:
        """Boolean to indicate index of the color channels in the sample `x`."""
        return self._channels_first

    @property
    def input_shape(self) -> tuple[int, int, int]:
        """Return the shape of one input sample.

        Returns:
            Shape of one input sample.
        """
        return cast(YOLOV1ObjectDetector, self._model).image_input_shape

    @property
    def native_label_is_pytorch_format(self) -> bool:
        return False

    def fit(self, x, y, **kwargs) -> None:
        raise NotImplementedError

    def predict(
        self,
        x: np.ndarray,
        batch_size: int = 16,
        standardise_output: bool = False,
        **kwargs,
    ) -> Any:
        """Perform prediction for a batch of inputs.

        Args:
            x: Samples of shape (nb_samples, height, width, nb_channels).
            batch_size: Batch size.
            standardise_output: True if output should be standardised to PyTorch format.
                Box coordinates will be scaled from [0, 1] to image dimensions, label
                index will be increased by 1 to adhere to COCO categories and the boxes
                will be changed to [x1, y1, x2, y2] format, with 0 <= x1 < x2 <= W and
                0 <= y1 < y2 <= H.

        Returns:
            Predictions of format `List[Dict[str, np.ndarray]]`, one for each input
            image. The fields of the Dict are as follows:
            - boxes [N, 4]: the boxes in [y1, x1, y2, x2] format, with 0 <= x1 < x2 <= W
                and 0 <= y1 < y2 <= H. Can be changed to PyTorch format with
                `standardise_output=True`.
            - labels [N]: the labels for each image in TensorFlow format. Can be changed
                to PyTorch format with `standardise_output=True`.
            - scores [N]: the scores or each prediction.
        """
        # Apply preprocessing
        x, _ = self._apply_preprocessing(x, y=None, fit=False)

        num_batches = math.ceil(x.shape[0] / batch_size)

        bbox_coordinates = []
        bbox_confidences = []
        class_probabilities = []

        for i_batch in range(num_batches):
            i_batch_start = i_batch * batch_size
            i_batch_end = min((i_batch + 1) * batch_size, x.shape[0])
            coord, conf, prob = self._model.predict(
                x=tf.convert_to_tensor(x[i_batch_start:i_batch_end]),
                **kwargs,
            )

            bbox_coordinates.append(coord)
            bbox_confidences.append(conf)
            class_probabilities.append(prob)

        predictions = self._bounding_boxes_postprocessing.postprocess(
            bboxes_cell_xywh=tf.concat(bbox_coordinates, axis=0),
            bboxes_conf=tf.concat(bbox_confidences, axis=0),
            bboxes_labels=tf.concat(class_probabilities, axis=0),
        )

        return self._encode_predict_output(
            predictions=predictions, standardise_output=standardise_output
        )

    def loss_gradient(
        self,
        x: np.ndarray,
        y: List[Dict[str, np.ndarray]],
        standardise_output: bool = False,
        **kwargs,
    ):
        """Compute the gradient of the loss function w.r.t. `x`.

        Args:
            x: Samples of shape (nb_samples, height, width, nb_channels).
            y: Targets of format `List[Dict[str, np.ndarray]]`, one for each input
                image. The fields of the Dict are as follows:
                - boxes [N, 4]: the boxes in [y1, x1, y2, x2] in scale [0, 1]
                    (`standardise_output=False`) or [x1, y1, x2, y2] in image scale
                    (`standardise_output=True`) format, with 0 <= x1 < x2 <= W and
                    0 <= y1 < y2 <= H.
                - labels [N]: the labels for each image in TensorFlow
                    (`standardise_output=False`) or PyTorch (`standardise_output=True`)
                    format
            standardise_output: True if `y` is provided in standardised PyTorch format.
                Box coordinates will be scaled back to [0, 1], label index will be
                decreased by 1 and the boxes will be changed from [x1, y1, x2, y2] to
                [y1, x1, y2, x2] format, with 0 <= x1 < x2 <= W and 0 <= y1 < y2 <= H.

        Returns:
            Loss gradients of the same shape as `x`.
        """
        # Apply preprocessing
        x_preprocessed, _ = self._apply_preprocessing(x, y=None, fit=False)

        x_tensor, y_tensor = self._decode_loss_gradient_input(
            x=x_preprocessed, y=y, standardise_output=standardise_output
        )
        input_var = tf.Variable(x_tensor)

        with GradientTape() as tape:
            tape.watch(input_var)
            y_pred = self._model(input_var, training=False)
            loss = self._model.loss(y_tensor, y_pred)

        grads = tape.gradient(loss, input_var).numpy()
        grads = self._apply_preprocessing_gradient(x, grads)

        return grads

    def _decode_loss_gradient_input(
        self, x: np.ndarray, y: List[Dict[str, np.ndarray]], standardise_output: bool
    ) -> Tuple[Tensor, Tuple[Tensor, Tensor, Tensor, Tensor]]:
        # Necessary to make copies of underlying numpy arrays, otherwise y mutates each
        # time this function is called with strandardise_output = True
        y = [{k: v.copy() for k, v in y_dict.items()} for y_dict in y]

        if standardise_output:
            y = convert_pt_to_tf(
                y=y, height=self.input_shape[0], width=self.input_shape[1]
            )

        bboxes_cell_xywh_grid = []
        bboxes_labels_grid = []
        bboxes_object_mask = []
        bboxes_no_object_mask = []

        for element in y:
            nonzero_scores_mask = element["scores"] > 0
            boxes = element["boxes"][..., [1, 0, 3, 2]][nonzero_scores_mask]
            labels = element["labels"][nonzero_scores_mask].astype("int32")

            y_embedded = self._bounding_boxes_postprocessing.embed(
                bboxes_corner=tf.convert_to_tensor(boxes, dtype=tf.float32),
                bboxes_labels=tf.convert_to_tensor(labels, dtype=tf.int32),
                n_classes=tf.convert_to_tensor(self._n_classes, dtype=tf.int32),
            )

            bboxes_cell_xywh_grid.append(y_embedded[0])
            bboxes_labels_grid.append(y_embedded[1])
            bboxes_object_mask.append(y_embedded[2])
            bboxes_no_object_mask.append(y_embedded[3])

        x_tensor = tf.convert_to_tensor(x)
        y_tensor = (
            tf.stack(bboxes_cell_xywh_grid, axis=0),
            tf.stack(bboxes_labels_grid, axis=0),
            tf.stack(bboxes_object_mask, axis=0),
            tf.stack(bboxes_no_object_mask, axis=0),
        )

        return x_tensor, y_tensor

    def _encode_predict_output(
        self,
        predictions: Tuple[Tensor, Tensor, Tensor, Tensor],
        standardise_output: bool,
    ) -> List[Dict[str, np.ndarray]]:
        batched_predictions = zip(
            predictions[0].numpy(),
            predictions[1].numpy(),
            predictions[2].numpy(),
        )
        results = [
            {"boxes": boxes.copy(), "labels": labels.copy(), "scores": scores.copy()}
            for boxes, scores, labels in batched_predictions
        ]

        if standardise_output:
            results = convert_tf_to_pt(
                y=results, height=self.input_shape[0], width=self.input_shape[1]
            )

        return results



"""Create tables for v1 schema

Revision ID: 148c032ffbd8
Revises: b835738272a8
Create Date: 2024-04-08 08:08:39.717348

"""

import sqlalchemy as sa
from alembic import op

from dioptra.restapi.db.custom_types import GUID, TZDateTime

# revision identifiers, used by Alembic.
revision = "148c032ffbd8"
down_revision = "b835738272a8"
branch_labels = None
depends_on = None

# Declare the database ontology
ENTRY_POINT_PARAMETER_TYPES = [
    {"parameter_type": "string"},
    {"parameter_type": "float"},
    {"parameter_type": "path"},
    {"parameter_type": "uri"},
]
JOB_STATUS_TYPES = [
    {"status": "queued"},
    {"status": "started"},
    {"status": "deferred"},
    {"status": "finished"},
    {"status": "failed"},
]
USER_LOCK_TYPES = [
    {"user_lock_type": "delete"},
]
GROUP_LOCK_TYPES = [
    {"group_lock_type": "delete"},
]
RESOURCE_LOCK_TYPES = [
    {"resource_lock_type": "delete"},
]
RESOURCE_TYPES = [
    {"resource_type": "artifact"},
    {"resource_type": "entry_point"},
    {"resource_type": "experiment"},
    {"resource_type": "job"},
    {"resource_type": "ml_model"},
    {"resource_type": "plugin"},
    {"resource_type": "plugin_file"},
    {"resource_type": "plugin_task_parameter_type"},
    {"resource_type": "queue"},
    {"resource_type": "resource_snapshot"},
]
RESOURCE_DEPENDENCY_TYPES = [
    {"parent_resource_type": "experiment", "child_resource_type": "entry_point"},
    {"parent_resource_type": "entry_point", "child_resource_type": "plugin"},
    {"parent_resource_type": "entry_point", "child_resource_type": "queue"},
    {"parent_resource_type": "plugin", "child_resource_type": "plugin_file"},
    {"parent_resource_type": "job", "child_resource_type": "artifact"},
    {"parent_resource_type": "job", "child_resource_type": "job"},
]
LEGACY_JOB_STATUS_TYPES = [
    {"status": "queued"},
    {"status": "started"},
    {"status": "deferred"},
    {"status": "finished"},
    {"status": "failed"},
]


# Helper function for inserting the ontology definitions
def _insert_ontology_data(
    entry_point_parameter_types_table,
    job_status_types_table,
    user_lock_types_table,
    group_lock_types_table,
    resource_lock_types_table,
    resource_types_table,
    resource_dependency_types_table,
) -> None:
    table_data_mappings = [
        (entry_point_parameter_types_table, ENTRY_POINT_PARAMETER_TYPES),
        (job_status_types_table, JOB_STATUS_TYPES),
        (user_lock_types_table, USER_LOCK_TYPES),
        (group_lock_types_table, GROUP_LOCK_TYPES),
        (resource_lock_types_table, RESOURCE_LOCK_TYPES),
        (resource_types_table, RESOURCE_TYPES),
        (resource_dependency_types_table, RESOURCE_DEPENDENCY_TYPES),
    ]

    for table, data in table_data_mappings:
        op.bulk_insert(table, data)


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    entry_point_parameter_types_table = op.create_table(
        "entry_point_parameter_types",
        sa.Column("parameter_type", sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint(
            "parameter_type", name=op.f("pk_entry_point_parameter_types")
        ),
    )
    group_lock_types_table = op.create_table(
        "group_lock_types",
        sa.Column("group_lock_type", sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint("group_lock_type", name=op.f("pk_group_lock_types")),
    )
    op.create_table(
        "job_mlflow_runs",
        sa.Column(
            "job_resource_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("mlflow_run_id", GUID(), nullable=False),
        sa.PrimaryKeyConstraint("job_resource_id", name=op.f("pk_job_mlflow_runs")),
    )
    with op.batch_alter_table("job_mlflow_runs", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_job_mlflow_runs_mlflow_run_id"),
            ["mlflow_run_id"],
            unique=True,
        )

    job_status_types_table = op.create_table(
        "job_status_types",
        sa.Column("status", sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint("status", name=op.f("pk_job_status_types")),
    )
    resource_lock_types_table = op.create_table(
        "resource_lock_types",
        sa.Column("resource_lock_type", sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint(
            "resource_lock_type", name=op.f("pk_resource_lock_types")
        ),
    )
    resource_types_table = op.create_table(
        "resource_types",
        sa.Column("resource_type", sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint("resource_type", name=op.f("pk_resource_types")),
    )
    user_lock_types_table = op.create_table(
        "user_lock_types",
        sa.Column("user_lock_type", sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint("user_lock_type", name=op.f("pk_user_lock_types")),
    )
    op.create_table(
        "users",
        sa.Column(
            "user_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("alternative_id", GUID(), nullable=False),
        sa.Column("username", sa.Text(), nullable=False),
        sa.Column("password", sa.Text(), nullable=True),
        sa.Column("email_address", sa.Text(), nullable=False),
        sa.Column("created_on", TZDateTime(), nullable=False),
        sa.Column("last_modified_on", TZDateTime(), nullable=False),
        sa.Column("last_login_on", TZDateTime(), nullable=True),
        sa.Column("password_expire_on", TZDateTime(), nullable=False),
        sa.PrimaryKeyConstraint("user_id", name=op.f("pk_users")),
    )
    with op.batch_alter_table("users", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_users_alternative_id"), ["alternative_id"], unique=True
        )
        batch_op.create_index(
            batch_op.f("ix_users_email_address"), ["email_address"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_users_username"), ["username"], unique=False
        )

    op.create_table(
        "groups",
        sa.Column(
            "group_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("name", sa.Text(), nullable=False),
        sa.Column(
            "user_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("created_on", TZDateTime(), nullable=False),
        sa.Column("last_modified_on", TZDateTime(), nullable=False),
        sa.PrimaryKeyConstraint("group_id", name=op.f("pk_groups")),
    )
    with op.batch_alter_table("groups", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_groups_user_id"), ["user_id"], unique=False
        )

    resource_dependency_types_table = op.create_table(
        "resource_dependency_types",
        sa.Column("parent_resource_type", sa.Text(), nullable=False),
        sa.Column("child_resource_type", sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint(
            "parent_resource_type",
            "child_resource_type",
            name=op.f("pk_resource_dependency_types"),
        ),
    )
    op.create_table(
        "user_locks",
        sa.Column(
            "user_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("user_lock_type", sa.Text(), nullable=False),
        sa.Column("created_on", TZDateTime(), nullable=False),
        sa.PrimaryKeyConstraint(
            "user_id", "user_lock_type", name=op.f("pk_user_locks")
        ),
    )
    op.create_table(
        "draft_resources",
        sa.Column(
            "draft_resource_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "group_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("resource_type", sa.Text(), nullable=False),
        sa.Column(
            "user_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("payload", sa.JSON(), nullable=False),
        sa.Column("created_on", TZDateTime(), nullable=False),
        sa.Column("last_modified_on", TZDateTime(), nullable=False),
        sa.PrimaryKeyConstraint("draft_resource_id", name=op.f("pk_draft_resources")),
    )
    with op.batch_alter_table("draft_resources", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_draft_resources_group_id"),
            ["group_id", "resource_type", "user_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_draft_resources_user_id"), ["user_id"], unique=False
        )

    op.create_table(
        "group_locks",
        sa.Column(
            "group_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("group_lock_type", sa.Text(), nullable=False),
        sa.Column("created_on", TZDateTime(), nullable=False),
        sa.PrimaryKeyConstraint(
            "group_id", "group_lock_type", name=op.f("pk_group_locks")
        ),
    )
    op.create_table(
        "group_managers",
        sa.Column(
            "user_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "group_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("owner", sa.Boolean(), nullable=False),
        sa.Column("admin", sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint("user_id", "group_id", name=op.f("pk_group_managers")),
    )
    op.create_table(
        "group_members",
        sa.Column(
            "user_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "group_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("read", sa.Boolean(), nullable=False),
        sa.Column("write", sa.Boolean(), nullable=False),
        sa.Column("share_read", sa.Boolean(), nullable=False),
        sa.Column("share_write", sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint("user_id", "group_id", name=op.f("pk_group_members")),
    )
    op.create_table(
        "resource_dependencies",
        sa.Column(
            "parent_resource_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "child_resource_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("parent_resource_type", sa.Text(), nullable=False),
        sa.Column("child_resource_type", sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint(
            "parent_resource_id",
            "child_resource_id",
            name=op.f("pk_resource_dependencies"),
        ),
    )
    with op.batch_alter_table("resource_dependencies", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_resource_dependencies_child_resource_id"),
            ["child_resource_id", "parent_resource_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_resource_dependencies_parent_resource_type"),
            ["parent_resource_type", "child_resource_type"],
            unique=False,
        )

    op.create_table(
        "resources",
        sa.Column(
            "resource_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "group_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("resource_type", sa.Text(), nullable=False),
        sa.Column("created_on", TZDateTime(), nullable=False),
        sa.PrimaryKeyConstraint("resource_id", name=op.f("pk_resources")),
    )
    with op.batch_alter_table("resources", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_resources_group_id"),
            ["group_id", "resource_type"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_resources_resource_id"),
            ["resource_id", "resource_type"],
            unique=True,
        )
        batch_op.create_index(
            batch_op.f("ix_resources_resource_type"), ["resource_type"], unique=False
        )

    op.create_table(
        "tags",
        sa.Column(
            "tag_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "group_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "user_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("name", sa.Text(), nullable=False),
        sa.Column("created_on", TZDateTime(), nullable=False),
        sa.Column("last_modified_on", TZDateTime(), nullable=False),
        sa.PrimaryKeyConstraint("tag_id", name=op.f("pk_tags")),
    )
    with op.batch_alter_table("tags", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_tags_group_id"), ["group_id", "name"], unique=True
        )

    op.create_table(
        "resource_locks",
        sa.Column(
            "resource_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("resource_lock_type", sa.Text(), nullable=False),
        sa.Column("created_on", TZDateTime(), nullable=False),
        sa.PrimaryKeyConstraint(
            "resource_id", "resource_lock_type", name=op.f("pk_resource_locks")
        ),
    )
    op.create_table(
        "resource_snapshots",
        sa.Column(
            "resource_snapshot_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "resource_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("resource_type", sa.Text(), nullable=False),
        sa.Column(
            "user_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("description", sa.Text(), nullable=True),
        sa.Column("created_on", TZDateTime(), nullable=False),
        sa.PrimaryKeyConstraint(
            "resource_snapshot_id", name=op.f("pk_resource_snapshots")
        ),
        sa.UniqueConstraint(
            "resource_snapshot_id",
            "resource_id",
            name=op.f("uq_resource_snapshots_resource_snapshot_id"),
        ),
    )
    with op.batch_alter_table("resource_snapshots", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_resource_snapshots_resource_id"),
            ["resource_id", "created_on"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_resource_snapshots_resource_snapshot_id"),
            ["resource_snapshot_id", "resource_id", "resource_type"],
            unique=True,
        )
        batch_op.create_index(
            batch_op.f("ix_resource_snapshots_resource_type"),
            ["resource_type"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_resource_snapshots_description"),
            ["description"],
            unique=False,
        )

    op.create_table(
        "resource_tags",
        sa.Column(
            "resource_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "tag_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.PrimaryKeyConstraint("resource_id", "tag_id", name=op.f("pk_resource_tags")),
    )
    op.create_table(
        "shared_resources",
        sa.Column(
            "shared_resource_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "resource_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "group_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "user_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("read", sa.Boolean(), nullable=False),
        sa.Column("write", sa.Boolean(), nullable=False),
        sa.Column("created_on", TZDateTime(), nullable=False),
        sa.PrimaryKeyConstraint("shared_resource_id", name=op.f("pk_shared_resources")),
    )
    with op.batch_alter_table("shared_resources", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_shared_resources_group_id"),
            ["group_id", "resource_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_shared_resources_resource_id"), ["resource_id"], unique=False
        )

    op.create_table(
        "artifacts",
        sa.Column(
            "resource_snapshot_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "resource_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("uri", sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint("resource_snapshot_id", name=op.f("pk_artifacts")),
    )
    with op.batch_alter_table("artifacts", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_artifacts_resource_id"), ["resource_id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_artifacts_resource_snapshot_id"),
            ["resource_snapshot_id", "resource_id"],
            unique=True,
        )

    op.create_table(
        "entry_points",
        sa.Column(
            "resource_snapshot_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "resource_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("name", sa.Text(), nullable=False),
        sa.Column("task_graph", sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint("resource_snapshot_id", name=op.f("pk_entry_points")),
    )
    with op.batch_alter_table("entry_points", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_entry_points_name"), ["name"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_entry_points_resource_id"), ["resource_id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_entry_points_resource_snapshot_id"),
            ["resource_snapshot_id", "resource_id"],
            unique=True,
        )

    op.create_table(
        "experiments",
        sa.Column(
            "resource_snapshot_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "resource_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("name", sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint("resource_snapshot_id", name=op.f("pk_experiments")),
    )
    with op.batch_alter_table("experiments", schema=None) as batch_op:
        batch_op.create_index(batch_op.f("ix_experiments_name"), ["name"], unique=False)
        batch_op.create_index(
            batch_op.f("ix_experiments_resource_id"), ["resource_id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_experiments_resource_snapshot_id"),
            ["resource_snapshot_id", "resource_id"],
            unique=True,
        )

    op.create_table(
        "jobs",
        sa.Column(
            "resource_snapshot_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "resource_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("timeout", sa.Text(), nullable=False),
        sa.Column("status", sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint("resource_snapshot_id", name=op.f("pk_jobs")),
    )
    with op.batch_alter_table("jobs", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_jobs_resource_id"), ["resource_id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_jobs_resource_snapshot_id"),
            ["resource_snapshot_id", "resource_id"],
            unique=True,
        )
        batch_op.create_index(batch_op.f("ix_jobs_status"), ["status"], unique=False)

    op.create_table(
        "ml_models",
        sa.Column(
            "resource_snapshot_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "resource_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "artifact_resource_snapshot_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("name", sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint("resource_snapshot_id", name=op.f("pk_ml_models")),
        sa.UniqueConstraint(
            "resource_snapshot_id",
            "resource_id",
            name=op.f("uq_ml_models_resource_snapshot_id"),
        ),
    )
    with op.batch_alter_table("ml_models", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_ml_models_resource_id"), ["resource_id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_ml_models_resource_snapshot_id"),
            ["resource_snapshot_id", "resource_id", "artifact_resource_snapshot_id"],
            unique=True,
        )

    op.create_table(
        "plugin_files",
        sa.Column(
            "resource_snapshot_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "resource_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("filename", sa.Text(), nullable=False),
        sa.Column("contents", sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint("resource_snapshot_id", name=op.f("pk_plugin_files")),
    )
    with op.batch_alter_table("plugin_files", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_plugin_files_filename"), ["filename"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_plugin_files_resource_id"), ["resource_id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_plugin_files_resource_snapshot_id"),
            ["resource_snapshot_id", "resource_id"],
            unique=True,
        )

    op.create_table(
        "plugin_task_parameter_types",
        sa.Column(
            "resource_snapshot_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "resource_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("name", sa.Text(), nullable=False),
        sa.Column("structure", sa.JSON(), nullable=True),
        sa.PrimaryKeyConstraint(
            "resource_snapshot_id", name=op.f("pk_plugin_task_parameter_types")
        ),
    )
    with op.batch_alter_table("plugin_task_parameter_types", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_plugin_task_parameter_types_name"), ["name"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_plugin_task_parameter_types_resource_id"),
            ["resource_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_plugin_task_parameter_types_resource_snapshot_id"),
            ["resource_snapshot_id", "resource_id"],
            unique=True,
        )

    op.create_table(
        "plugins",
        sa.Column(
            "resource_snapshot_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "resource_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("name", sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint("resource_snapshot_id", name=op.f("pk_plugins")),
    )
    with op.batch_alter_table("plugins", schema=None) as batch_op:
        batch_op.create_index(batch_op.f("ix_plugins_name"), ["name"], unique=False)
        batch_op.create_index(
            batch_op.f("ix_plugins_resource_id"), ["resource_id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_plugins_resource_snapshot_id"),
            ["resource_snapshot_id", "resource_id"],
            unique=True,
        )

    op.create_table(
        "queues",
        sa.Column(
            "resource_snapshot_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "resource_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("name", sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint("resource_snapshot_id", name=op.f("pk_queues")),
    )
    with op.batch_alter_table("queues", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_queues_resource_id"), ["resource_id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_queues_resource_snapshot_id"),
            ["resource_snapshot_id", "resource_id"],
            unique=True,
        )

    op.create_table(
        "shared_resource_tags",
        sa.Column(
            "shared_resource_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "tag_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.PrimaryKeyConstraint(
            "shared_resource_id", "tag_id", name=op.f("pk_shared_resource_tags")
        ),
    )
    op.create_table(
        "entry_point_jobs",
        sa.Column(
            "entry_point_resource_snapshot_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "job_resource_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.PrimaryKeyConstraint(
            "entry_point_resource_snapshot_id",
            "job_resource_id",
            name=op.f("pk_entry_point_jobs"),
        ),
    )
    with op.batch_alter_table("entry_point_jobs", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_entry_point_jobs_job_resource_id"),
            ["job_resource_id"],
            unique=True,
        )

    op.create_table(
        "entry_point_parameters",
        sa.Column(
            "entry_point_resource_snapshot_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "parameter_number",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("parameter_type", sa.Text(), nullable=False),
        sa.Column("name", sa.Text(), nullable=False),
        sa.Column("default_value", sa.Text(), nullable=True),
        sa.ForeignKeyConstraint(
            ["entry_point_resource_snapshot_id"],
            ["entry_points.resource_snapshot_id"],
            name=op.f(
                "fk_entry_point_parameters_entry_point_resource_snapshot_id_entry_points"
            ),
        ),
        sa.ForeignKeyConstraint(
            ["parameter_type"],
            ["entry_point_parameter_types.parameter_type"],
            name=op.f(
                "fk_entry_point_parameters_parameter_type_entry_point_parameter_types"
            ),
        ),
        sa.PrimaryKeyConstraint(
            "entry_point_resource_snapshot_id",
            "parameter_number",
            name=op.f("pk_entry_point_parameters"),
        ),
    )
    with op.batch_alter_table("entry_point_parameters", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_entry_point_parameters_entry_point_resource_snapshot_id"),
            ["entry_point_resource_snapshot_id", "name"],
            unique=True,
        )
        batch_op.create_index(
            batch_op.f("ix_entry_point_parameters_parameter_type"),
            ["parameter_type"],
            unique=False,
        )

    op.create_table(
        "entry_point_plugin_files",
        sa.Column(
            "entry_point_resource_snapshot_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "plugin_resource_snapshot_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "plugin_file_resource_snapshot_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.PrimaryKeyConstraint(
            "entry_point_resource_snapshot_id",
            "plugin_resource_snapshot_id",
            "plugin_file_resource_snapshot_id",
            name=op.f("pk_entry_point_plugin_files"),
        ),
    )
    op.create_table(
        "experiment_jobs",
        sa.Column(
            "experiment_resource_snapshot_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "job_resource_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.PrimaryKeyConstraint(
            "experiment_resource_snapshot_id",
            "job_resource_id",
            name=op.f("pk_experiment_jobs"),
        ),
    )
    with op.batch_alter_table("experiment_jobs", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_experiment_jobs_job_resource_id"),
            ["job_resource_id"],
            unique=True,
        )

    op.create_table(
        "plugin_tasks",
        sa.Column(
            "plugin_file_resource_snapshot_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("plugin_task_name", sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint(
            "plugin_file_resource_snapshot_id",
            "plugin_task_name",
            name=op.f("pk_plugin_tasks"),
        ),
    )
    op.create_table(
        "queue_jobs",
        sa.Column(
            "queue_resource_snapshot_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "job_resource_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.PrimaryKeyConstraint(
            "queue_resource_snapshot_id", "job_resource_id", name=op.f("pk_queue_jobs")
        ),
    )
    with op.batch_alter_table("queue_jobs", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_queue_jobs_job_resource_id"),
            ["job_resource_id"],
            unique=True,
        )

    op.create_table(
        "entry_point_parameter_values",
        sa.Column(
            "job_resource_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "entry_point_resource_snapshot_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "parameter_number",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("value", sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint(
            "job_resource_id",
            "entry_point_resource_snapshot_id",
            "parameter_number",
            name=op.f("pk_entry_point_parameter_values"),
        ),
    )
    with op.batch_alter_table("entry_point_parameter_values", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_entry_point_parameter_values_job_resource_id"),
            ["job_resource_id", "parameter_number"],
            unique=True,
        )

    op.create_table(
        "plugin_task_input_parameters",
        sa.Column(
            "plugin_file_resource_snapshot_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("plugin_task_name", sa.Text(), nullable=False),
        sa.Column(
            "parameter_number",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "plugin_task_parameter_type_resource_snapshot_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("name", sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint(
            "plugin_file_resource_snapshot_id",
            "plugin_task_name",
            "parameter_number",
            name=op.f("pk_plugin_task_input_parameters"),
        ),
    )
    with op.batch_alter_table("plugin_task_input_parameters", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f(
                "ix_plugin_task_input_parameters_plugin_file_resource_snapshot_id"
            ),
            ["plugin_file_resource_snapshot_id", "plugin_task_name"],
            unique=True,
        )
        batch_op.create_index(
            batch_op.f(
                "ix_plugin_task_input_parameters_plugin_task_parameter_type_resource_snapshot_id"
            ),
            ["plugin_task_parameter_type_resource_snapshot_id"],
            unique=False,
        )

    op.create_table(
        "plugin_task_output_parameters",
        sa.Column(
            "plugin_file_resource_snapshot_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("plugin_task_name", sa.Text(), nullable=False),
        sa.Column(
            "parameter_number",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column(
            "plugin_task_parameter_type_resource_snapshot_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("name", sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint(
            "plugin_file_resource_snapshot_id",
            "plugin_task_name",
            "parameter_number",
            name=op.f("pk_plugin_task_output_parameters"),
        ),
    )
    with op.batch_alter_table("plugin_task_output_parameters", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f(
                "ix_plugin_task_output_parameters_plugin_file_resource_snapshot_id"
            ),
            ["plugin_file_resource_snapshot_id", "plugin_task_name"],
            unique=True,
        )
        batch_op.create_index(
            batch_op.f(
                "ix_plugin_task_output_parameters_plugin_task_parameter_type_resource_snapshot_id"
            ),
            ["plugin_task_parameter_type_resource_snapshot_id"],
            unique=False,
        )

    # Apply foreign keys after creating tables and indexes
    with op.batch_alter_table("groups", schema=None) as batch_op:
        batch_op.create_foreign_key(
            "fk_groups_user_id_users", "users", ["user_id"], ["user_id"]
        )

    with op.batch_alter_table("resource_dependency_types", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f(
                "fk_resource_dependency_types_parent_resource_type_resource_types"
            ),
            "resource_types",
            ["parent_resource_type"],
            ["resource_type"],
        )
        batch_op.create_foreign_key(
            batch_op.f(
                "fk_resource_dependency_types_child_resource_type_resource_types"
            ),
            "resource_types",
            ["child_resource_type"],
            ["resource_type"],
        )

    with op.batch_alter_table("user_locks", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f("fk_user_locks_user_id_users"), "users", ["user_id"], ["user_id"]
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_user_locks_user_lock_type_user_lock_types"),
            "user_lock_types",
            ["user_lock_type"],
            ["user_lock_type"],
        )

    with op.batch_alter_table("draft_resources", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f("fk_draft_resources_group_id_groups"),
            "groups",
            ["group_id"],
            ["group_id"],
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_draft_resources_resource_type_resource_types"),
            "resource_types",
            ["resource_type"],
            ["resource_type"],
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_draft_resources_user_id_users"),
            "users",
            ["user_id"],
            ["user_id"],
        )

    with op.batch_alter_table("group_locks", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f("fk_group_locks_group_id_groups"),
            "groups",
            ["group_id"],
            ["group_id"],
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_group_locks_group_lock_type_group_lock_types"),
            "group_lock_types",
            ["group_lock_type"],
            ["group_lock_type"],
        )

    with op.batch_alter_table("group_managers", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f("fk_group_managers_group_id_groups"),
            "groups",
            ["group_id"],
            ["group_id"],
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_group_managers_user_id_users"),
            "users",
            ["user_id"],
            ["user_id"],
        )

    with op.batch_alter_table("group_members", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f("fk_group_members_group_id_groups"),
            "groups",
            ["group_id"],
            ["group_id"],
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_group_members_user_id_users"),
            "users",
            ["user_id"],
            ["user_id"],
        )

    with op.batch_alter_table("resource_dependencies", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f(
                "fk_resource_dependencies_parent_resource_type_resource_dependency_types"
            ),
            "resource_dependency_types",
            ["parent_resource_type", "child_resource_type"],
            ["parent_resource_type", "child_resource_type"],
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_resource_dependencies_parent_resource_id_resources"),
            "resources",
            ["parent_resource_id", "parent_resource_type"],
            ["resource_id", "resource_type"],
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_resource_dependencies_child_resource_id_resources"),
            "resources",
            ["child_resource_id", "child_resource_type"],
            ["resource_id", "resource_type"],
        )

    with op.batch_alter_table("resources", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f("fk_resources_group_id_groups"),
            "groups",
            ["group_id"],
            ["group_id"],
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_resources_resource_type_resource_types"),
            "resource_types",
            ["resource_type"],
            ["resource_type"],
        )

    with op.batch_alter_table("tags", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f("fk_tags_group_id_groups"), "groups", ["group_id"], ["group_id"]
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_tags_user_id_users"), "users", ["user_id"], ["user_id"]
        )

    with op.batch_alter_table("resource_locks", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f("fk_resource_locks_resource_id_resources"),
            "resources",
            ["resource_id"],
            ["resource_id"],
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_resource_locks_resource_lock_type_resource_lock_types"),
            "resource_lock_types",
            ["resource_lock_type"],
            ["resource_lock_type"],
        )

    with op.batch_alter_table("resource_snapshots", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f("fk_resource_snapshots_resource_id_resources"),
            "resources",
            ["resource_id", "resource_type"],
            ["resource_id", "resource_type"],
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_resource_snapshots_resource_type_resource_types"),
            "resource_types",
            ["resource_type"],
            ["resource_type"],
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_resource_snapshots_user_id_users"),
            "users",
            ["user_id"],
            ["user_id"],
        )

    with op.batch_alter_table("resource_tags", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f("fk_resource_tags_resource_id_resources"),
            "resources",
            ["resource_id"],
            ["resource_id"],
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_resource_tags_tag_id_tags"), "tags", ["tag_id"], ["tag_id"]
        )

    with op.batch_alter_table("shared_resources", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f("fk_shared_resources_group_id_groups"),
            "groups",
            ["group_id"],
            ["group_id"],
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_shared_resources_resource_id_resources"),
            "resources",
            ["resource_id"],
            ["resource_id"],
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_shared_resources_user_id_users"),
            "users",
            ["user_id"],
            ["user_id"],
        )

    with op.batch_alter_table("artifacts", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f("fk_artifacts_resource_snapshot_id_resource_snapshots"),
            "resource_snapshots",
            ["resource_snapshot_id", "resource_id"],
            ["resource_snapshot_id", "resource_id"],
        )

    with op.batch_alter_table("entry_points", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f("fk_entry_points_resource_snapshot_id_resource_snapshots"),
            "resource_snapshots",
            ["resource_snapshot_id", "resource_id"],
            ["resource_snapshot_id", "resource_id"],
        )

    with op.batch_alter_table("experiments", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f("fk_experiments_resource_snapshot_id_resource_snapshots"),
            "resource_snapshots",
            ["resource_snapshot_id", "resource_id"],
            ["resource_snapshot_id", "resource_id"],
        )

    with op.batch_alter_table("jobs", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f("fk_jobs_resource_snapshot_id_resource_snapshots"),
            "resource_snapshots",
            ["resource_snapshot_id", "resource_id"],
            ["resource_snapshot_id", "resource_id"],
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_jobs_status_job_status_types"),
            "job_status_types",
            ["status"],
            ["status"],
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_jobs_resource_id_entry_point_jobs"),
            "entry_point_jobs",
            ["resource_id"],
            ["job_resource_id"],
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_jobs_resource_id_experiment_jobs"),
            "experiment_jobs",
            ["resource_id"],
            ["job_resource_id"],
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_jobs_resource_id_queue_jobs"),
            "queue_jobs",
            ["resource_id"],
            ["job_resource_id"],
        )

    with op.batch_alter_table("ml_models", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f("fk_ml_models_resource_snapshot_id_resource_snapshots"),
            "resource_snapshots",
            ["resource_snapshot_id", "resource_id"],
            ["resource_snapshot_id", "resource_id"],
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_ml_models_artifact_resource_snapshot_id_artifacts"),
            "artifacts",
            ["artifact_resource_snapshot_id"],
            ["resource_snapshot_id"],
        )

    with op.batch_alter_table("plugin_files", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f("fk_plugin_files_resource_snapshot_id_resource_snapshots"),
            "resource_snapshots",
            ["resource_snapshot_id", "resource_id"],
            ["resource_snapshot_id", "resource_id"],
        )

    with op.batch_alter_table("plugin_task_parameter_types", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f(
                "fk_plugin_task_parameter_types_resource_snapshot_id_resource_snapshots"
            ),
            "resource_snapshots",
            ["resource_snapshot_id", "resource_id"],
            ["resource_snapshot_id", "resource_id"],
        )

    with op.batch_alter_table("plugins", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f("fk_plugins_resource_snapshot_id_resource_snapshots"),
            "resource_snapshots",
            ["resource_snapshot_id", "resource_id"],
            ["resource_snapshot_id", "resource_id"],
        )

    with op.batch_alter_table("queues", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f("fk_queues_resource_snapshot_id_resource_snapshots"),
            "resource_snapshots",
            ["resource_snapshot_id", "resource_id"],
            ["resource_snapshot_id", "resource_id"],
        )

    with op.batch_alter_table("shared_resource_tags", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f("fk_shared_resource_tags_shared_resource_id_shared_resources"),
            "shared_resources",
            ["shared_resource_id"],
            ["shared_resource_id"],
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_shared_resource_tags_tag_id_tags"),
            "tags",
            ["tag_id"],
            ["tag_id"],
        )

    with op.batch_alter_table("entry_point_jobs", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f(
                "fk_entry_point_jobs_entry_point_resource_snapshot_id_entry_points"
            ),
            "entry_points",
            ["entry_point_resource_snapshot_id"],
            ["resource_snapshot_id"],
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_entry_point_jobs_job_resource_id_resources"),
            "resources",
            ["job_resource_id"],
            ["resource_id"],
        )

    with op.batch_alter_table("entry_point_plugin_files", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f(
                "fk_entry_point_plugin_files_entry_point_resource_snapshot_id_entry_points"
            ),
            "entry_points",
            ["entry_point_resource_snapshot_id"],
            ["resource_snapshot_id"],
        )
        batch_op.create_foreign_key(
            batch_op.f(
                "fk_entry_point_plugin_files_plugin_file_resource_snapshot_id_plugin_files"
            ),
            "plugin_files",
            ["plugin_file_resource_snapshot_id"],
            ["resource_snapshot_id"],
        )
        batch_op.create_foreign_key(
            batch_op.f(
                "fk_entry_point_plugin_files_plugin_resource_snapshot_id_plugins"
            ),
            "plugins",
            ["plugin_resource_snapshot_id"],
            ["resource_snapshot_id"],
        )

    with op.batch_alter_table("experiment_jobs", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f(
                "fk_experiment_jobs_experiment_resource_snapshot_id_experiments"
            ),
            "experiments",
            ["experiment_resource_snapshot_id"],
            ["resource_snapshot_id"],
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_experiment_jobs_job_resource_id_resources"),
            "resources",
            ["job_resource_id"],
            ["resource_id"],
        )

    with op.batch_alter_table("job_mlflow_runs", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f("fk_job_mlflow_runs_job_resource_id_resources"),
            "resources",
            ["job_resource_id"],
            ["resource_id"],
        )

    with op.batch_alter_table("plugin_tasks", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f("fk_plugin_tasks_plugin_file_resource_snapshot_id_plugin_files"),
            "plugin_files",
            ["plugin_file_resource_snapshot_id"],
            ["resource_snapshot_id"],
        )

    with op.batch_alter_table("queue_jobs", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f("fk_queue_jobs_queue_resource_snapshot_id_queues"),
            "queues",
            ["queue_resource_snapshot_id"],
            ["resource_snapshot_id"],
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_queue_jobs_job_resource_id_resources"),
            "resources",
            ["job_resource_id"],
            ["resource_id"],
        )

    with op.batch_alter_table("entry_point_parameter_values", schema=None) as batch_op:
        batch_op.create_foreign_key(
            op.f(
                "fk_entry_point_parameter_values_entry_point_resource_snapshot_id_entry_point_parameters"
            ),
            "entry_point_parameters",
            ["entry_point_resource_snapshot_id", "parameter_number"],
            ["entry_point_resource_snapshot_id", "parameter_number"],
        )
        batch_op.create_foreign_key(
            op.f(
                "fk_entry_point_parameter_values_entry_point_resource_snapshot_id_entry_point_jobs"
            ),
            "entry_point_jobs",
            ["entry_point_resource_snapshot_id", "job_resource_id"],
            ["entry_point_resource_snapshot_id", "job_resource_id"],
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_entry_point_parameter_values_job_resource_id_resources"),
            "resources",
            ["job_resource_id"],
            ["resource_id"],
        )

    with op.batch_alter_table("plugin_task_input_parameters", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f(
                "fk_plugin_task_input_parameters_plugin_file_resource_snapshot_id_plugin_tasks"
            ),
            "plugin_tasks",
            ["plugin_file_resource_snapshot_id", "plugin_task_name"],
            ["plugin_file_resource_snapshot_id", "plugin_task_name"],
        )
        batch_op.create_foreign_key(
            batch_op.f(
                "fk_plugin_task_input_parameters_plugin_task_parameter_type_resource_snapshot_id_plugin_task_parameter_types"
            ),
            "plugin_task_parameter_types",
            ["plugin_task_parameter_type_resource_snapshot_id"],
            ["resource_snapshot_id"],
        )

    with op.batch_alter_table("plugin_task_output_parameters", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f(
                "fk_plugin_task_output_parameters_plugin_file_resource_snapshot_id_plugin_tasks"
            ),
            "plugin_tasks",
            ["plugin_file_resource_snapshot_id", "plugin_task_name"],
            ["plugin_file_resource_snapshot_id", "plugin_task_name"],
        )
        batch_op.create_foreign_key(
            batch_op.f(
                "fk_plugin_task_output_parameters_plugin_task_parameter_type_resource_snapshot_id_plugin_task_parameter_types"
            ),
            "plugin_task_parameter_types",
            ["plugin_task_parameter_type_resource_snapshot_id"],
            ["resource_snapshot_id"],
        )

    # Insert the ontology data
    _insert_ontology_data(
        entry_point_parameter_types_table=entry_point_parameter_types_table,
        job_status_types_table=job_status_types_table,
        user_lock_types_table=user_lock_types_table,
        group_lock_types_table=group_lock_types_table,
        resource_lock_types_table=resource_lock_types_table,
        resource_types_table=resource_types_table,
        resource_dependency_types_table=resource_dependency_types_table,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("plugin_task_output_parameters", schema=None) as batch_op:
        batch_op.drop_index(
            batch_op.f(
                "ix_plugin_task_output_parameters_plugin_task_parameter_type_resource_snapshot_id"
            )
        )
        batch_op.drop_index(
            batch_op.f(
                "ix_plugin_task_output_parameters_plugin_file_resource_snapshot_id"
            )
        )

    op.drop_table("plugin_task_output_parameters")
    with op.batch_alter_table("plugin_task_input_parameters", schema=None) as batch_op:
        batch_op.drop_index(
            batch_op.f(
                "ix_plugin_task_input_parameters_plugin_task_parameter_type_resource_snapshot_id"
            )
        )
        batch_op.drop_index(
            batch_op.f(
                "ix_plugin_task_input_parameters_plugin_file_resource_snapshot_id"
            )
        )

    op.drop_table("plugin_task_input_parameters")
    with op.batch_alter_table("entry_point_parameter_values", schema=None) as batch_op:
        batch_op.drop_index(
            batch_op.f("ix_entry_point_parameter_values_job_resource_id")
        )

    op.drop_table("entry_point_parameter_values")
    op.drop_table("plugin_tasks")
    op.drop_table("entry_point_plugin_files")
    with op.batch_alter_table("entry_point_parameters", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_entry_point_parameters_parameter_type"))
        batch_op.drop_index(
            batch_op.f("ix_entry_point_parameters_entry_point_resource_snapshot_id")
        )

    op.drop_table("entry_point_parameters")
    op.drop_table("shared_resource_tags")
    with op.batch_alter_table("queues", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_queues_resource_snapshot_id"))
        batch_op.drop_index(batch_op.f("ix_queues_resource_id"))

    with op.batch_alter_table("jobs", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_jobs_status"))
        batch_op.drop_index(batch_op.f("ix_jobs_resource_snapshot_id"))
        batch_op.drop_index(batch_op.f("ix_jobs_resource_id"))

    op.drop_table("jobs")
    with op.batch_alter_table("queue_jobs", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_queue_jobs_job_resource_id"))

    op.drop_table("queue_jobs")
    with op.batch_alter_table("experiment_jobs", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_experiment_jobs_job_resource_id"))

    op.drop_table("experiment_jobs")
    with op.batch_alter_table("entry_point_jobs", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_entry_point_jobs_job_resource_id"))

    op.drop_table("entry_point_jobs")
    with op.batch_alter_table("job_mlflow_runs", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_job_mlflow_runs_mlflow_run_id"))

    op.drop_table("job_mlflow_runs")
    op.drop_table("queues")
    with op.batch_alter_table("plugins", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_plugins_resource_snapshot_id"))
        batch_op.drop_index(batch_op.f("ix_plugins_resource_id"))
        batch_op.drop_index(batch_op.f("ix_plugins_name"))

    op.drop_table("plugins")
    with op.batch_alter_table("plugin_task_parameter_types", schema=None) as batch_op:
        batch_op.drop_index(
            batch_op.f("ix_plugin_task_parameter_types_resource_snapshot_id")
        )
        batch_op.drop_index(batch_op.f("ix_plugin_task_parameter_types_resource_id"))
        batch_op.drop_index(batch_op.f("ix_plugin_task_parameter_types_name"))

    op.drop_table("plugin_task_parameter_types")
    with op.batch_alter_table("plugin_files", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_plugin_files_resource_snapshot_id"))
        batch_op.drop_index(batch_op.f("ix_plugin_files_resource_id"))
        batch_op.drop_index(batch_op.f("ix_plugin_files_filename"))

    op.drop_table("plugin_files")
    with op.batch_alter_table("ml_models", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_ml_models_resource_snapshot_id"))
        batch_op.drop_index(batch_op.f("ix_ml_models_resource_id"))

    op.drop_table("ml_models")
    with op.batch_alter_table("experiments", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_experiments_resource_snapshot_id"))
        batch_op.drop_index(batch_op.f("ix_experiments_resource_id"))
        batch_op.drop_index(batch_op.f("ix_experiments_name"))

    op.drop_table("experiments")
    with op.batch_alter_table("entry_points", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_entry_points_resource_snapshot_id"))
        batch_op.drop_index(batch_op.f("ix_entry_points_resource_id"))
        batch_op.drop_index(batch_op.f("ix_entry_points_name"))

    op.drop_table("entry_points")
    with op.batch_alter_table("artifacts", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_artifacts_resource_snapshot_id"))
        batch_op.drop_index(batch_op.f("ix_artifacts_resource_id"))

    op.drop_table("artifacts")
    with op.batch_alter_table("shared_resources", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_shared_resources_resource_id"))
        batch_op.drop_index(batch_op.f("ix_shared_resources_group_id"))

    op.drop_table("shared_resources")
    op.drop_table("resource_tags")
    with op.batch_alter_table("resource_snapshots", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_resource_snapshots_description"))
        batch_op.drop_index(batch_op.f("ix_resource_snapshots_resource_type"))
        batch_op.drop_index(batch_op.f("ix_resource_snapshots_resource_snapshot_id"))
        batch_op.drop_index(batch_op.f("ix_resource_snapshots_resource_id"))

    op.drop_table("resource_snapshots")
    op.drop_table("resource_locks")
    with op.batch_alter_table("tags", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_tags_group_id"))

    op.drop_table("tags")
    with op.batch_alter_table("resource_dependencies", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_resource_dependencies_parent_resource_type"))
        batch_op.drop_index(batch_op.f("ix_resource_dependencies_child_resource_id"))

    op.drop_table("resource_dependencies")
    with op.batch_alter_table("resources", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_resources_resource_type"))
        batch_op.drop_index(batch_op.f("ix_resources_resource_id"))
        batch_op.drop_index(batch_op.f("ix_resources_group_id"))

    op.drop_table("resources")
    op.drop_table("group_members")
    op.drop_table("group_managers")
    op.drop_table("group_locks")
    with op.batch_alter_table("draft_resources", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_draft_resources_user_id"))
        batch_op.drop_index(batch_op.f("ix_draft_resources_group_id"))

    op.drop_table("draft_resources")
    op.drop_table("user_locks")
    op.drop_table("resource_dependency_types")
    with op.batch_alter_table("groups", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_groups_user_id"))

    op.drop_table("groups")
    with op.batch_alter_table("users", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_users_username"))
        batch_op.drop_index(batch_op.f("ix_users_email_address"))
        batch_op.drop_index(batch_op.f("ix_users_alternative_id"))

    op.drop_table("users")
    op.drop_table("user_lock_types")
    op.drop_table("resource_types")
    op.drop_table("resource_lock_types")
    op.drop_table("job_status_types")
    op.drop_table("group_lock_types")
    op.drop_table("entry_point_parameter_types")
    # ### end Alembic commands ###



"""Add 'readonly' resource lock type to db ontology

Revision ID: 10f9e72e72aa
Revises: 5b80d059cbb4
Create Date: 2024-06-28 10:12:34.014151

"""

from typing import Annotated

import sqlalchemy as sa
from alembic import op
from sqlalchemy.orm import (
    DeclarativeBase,
    Mapped,
    MappedAsDataclass,
    mapped_column,
    sessionmaker,
)

# revision identifiers, used by Alembic.
revision = "10f9e72e72aa"
down_revision = "5b80d059cbb4"
branch_labels = None
depends_on = None


# Migration data models
text_ = Annotated[str, mapped_column(sa.Text())]


class UpgradeBase(DeclarativeBase, MappedAsDataclass):
    pass


class ResourceLockTypeUpgrade(UpgradeBase):
    __tablename__ = "resource_lock_types"

    resource_lock_type: Mapped[text_] = mapped_column(primary_key=True)


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    bind = op.get_bind()
    Session = sessionmaker(bind=bind)

    with Session() as session:
        # Check whether the readonly lock type is already in the database definition
        # (this can happen if you upgrade then downgrade this migration).
        readonly_lock_type_stmt = sa.select(ResourceLockTypeUpgrade).where(
            ResourceLockTypeUpgrade.resource_lock_type == "readonly"
        )

        if session.scalar(readonly_lock_type_stmt) is None:
            # Add "readonly" to lock types table if its missing.
            session.add(ResourceLockTypeUpgrade(resource_lock_type="readonly"))

        session.commit()


def downgrade():
    # No downgrade necessary, the readonly entry in the resource_lock_types table
    # doesn't affect any other functionality.
    pass



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""Fixtures representing resources needed for test suites"""
import textwrap
from collections.abc import Iterator
from pathlib import Path
from typing import Any, cast

import pytest
from flask import Flask
from flask.testing import FlaskClient
from flask_sqlalchemy import SQLAlchemy
from injector import Injector
from pytest import MonkeyPatch

from ..lib import actions, mock_rq


@pytest.fixture
def app(dependency_modules: list[Any]) -> Iterator[Flask]:
    from dioptra.restapi import create_app

    injector = Injector(dependency_modules)
    app = create_app(env="test_v1", injector=injector)

    yield app


@pytest.fixture
def registered_users(client: FlaskClient, db: SQLAlchemy) -> dict[str, Any]:
    password = "supersecurepassword"
    user1_response = actions.register_user(
        client, "user1", "user1@example.org", password
    ).get_json()
    user2_response = actions.register_user(
        client, "user2", "user2@example.net", password
    ).get_json()
    user3_response = actions.register_user(
        client, "name", "user3@example.org", password
    ).get_json()
    users_info = {
        "user1": user1_response,
        "user2": user2_response,
        "user3": user3_response,
    }

    for _, user_info in users_info.items():
        user_info["password"] = password
        user_info["default_group_id"] = user_info["groups"][0]["id"]

    return users_info


@pytest.fixture
def auth_account(
    client: FlaskClient,
    db: SQLAlchemy,
    registered_users: dict[str, Any],  # noqa: F811
) -> dict[str, Any]:
    user_info = cast(dict[str, Any], registered_users["user1"])
    login_response = actions.login(
        client, username=user_info["username"], password=user_info["password"]
    )
    if login_response.status_code != 200:
        raise ValueError("User login failed.")
    return user_info


@pytest.fixture
def registered_tags(
    client: FlaskClient, db: SQLAlchemy, auth_account: dict[str, Any]
) -> dict[str, Any]:
    tag1_response = actions.register_tag(
        client,
        name="tag_one",
        group_id=auth_account["default_group_id"],
    ).get_json()
    tag2_response = actions.register_tag(
        client,
        name="tag_two",
        group_id=auth_account["default_group_id"],
    ).get_json()
    tag3_response = actions.register_tag(
        client,
        name="name",
        group_id=auth_account["default_group_id"],
    ).get_json()
    return {
        "tag1": tag1_response,
        "tag2": tag2_response,
        "tag3": tag3_response,
    }


@pytest.fixture
def registered_artifacts(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_jobs: dict[str, Any],
) -> dict[str, Any]:
    group_id = auth_account["groups"][0]["id"]
    artifact1_response = actions.register_artifact(
        client,
        uri="s3://bucket/model_v1.artifact",
        description="Model artifact.",
        job_id=registered_jobs["job1"]["id"],
        group_id=group_id,
    ).get_json()
    artifact2_response = actions.register_artifact(
        client,
        uri="s3://bucket/cnn.artifact",
        description="Trained conv net model artifact.",
        job_id=registered_jobs["job1"]["id"],
        group_id=group_id,
    ).get_json()
    artifact3_response = actions.register_artifact(
        client,
        uri="s3://bucket/model.artifact",
        description="Another model",
        job_id=registered_jobs["job2"]["id"],
        group_id=group_id,
    ).get_json()
    artifact4_response = actions.register_artifact(
        client,
        uri="s3://bucket/model_v2.artifact",
        description="Fine-tuned model.",
        job_id=registered_jobs["job3"]["id"],
        group_id=group_id,
    ).get_json()

    return {
        "artifact1": artifact1_response,
        "artifact2": artifact2_response,
        "artifact3": artifact3_response,
        "artifact4": artifact4_response,
    }


@pytest.fixture
def registered_models(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_artifacts: dict[str, Any],
) -> dict[str, Any]:
    group_id = auth_account["groups"][0]["id"]
    model1_response = actions.register_model(
        client,
        name="my_tensorflow_model",
        description="Trained model",
        group_id=group_id,
    ).get_json()
    model2_response = actions.register_model(
        client,
        name="model2",
        description="Trained model",
        group_id=group_id,
    ).get_json()
    model3_response = actions.register_model(
        client,
        name="model3",
        description="",
        group_id=group_id,
    ).get_json()

    # actions.register_model_version(
    #     client,
    #     model_id=model1_response["id"],
    #     artifact_id=registered_artifacts["artifact1"]["id"],
    #     description="initial version",
    # ).get_json()
    # actions.register_model_version(
    #     client,
    #     model_id=model2_response["id"],
    #     artifact_id=registered_artifacts["artifact2"]["id"],
    #     description="initial version",
    # ).get_json()
    # actions.register_model_version(
    #     client,
    #     model_id=model3_response["id"],
    #     artifact_id=registered_artifacts["artifact3"]["id"],
    # ).get_json()
    # actions.register_model_version(
    #     client,
    #     model_id=model1_response["id"],
    #     artifact_id=registered_artifacts["artifact4"]["id"],
    #     description="new version",
    # ).get_json()

    # return {
    #     "model1": actions.get_model(client, model1_response["id"]).get_json(),
    #     "model2": actions.get_model(client, model2_response["id"]).get_json(),
    #     "model3": actions.get_model(client, model3_response["id"]).get_json(),
    # }

    return {
        "model1": model1_response,
        "model2": model2_response,
        "model3": model3_response,
    }


@pytest.fixture
def registered_model_versions(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_models: dict[str, Any],
    registered_artifacts: dict[str, Any],
) -> dict[str, Any]:
    version1_response = actions.register_model_version(
        client,
        model_id=registered_models["model1"]["id"],
        artifact_id=registered_artifacts["artifact1"]["id"],
        description="initial version",
    ).get_json()
    version2_response = actions.register_model_version(
        client,
        model_id=registered_models["model2"]["id"],
        artifact_id=registered_artifacts["artifact2"]["id"],
        description="initial version",
    ).get_json()
    version3_response = actions.register_model_version(
        client,
        model_id=registered_models["model3"]["id"],
        artifact_id=registered_artifacts["artifact3"]["id"],
        description="initial version",
    ).get_json()
    version4_response = actions.register_model_version(
        client,
        model_id=registered_models["model2"]["id"],
        artifact_id=registered_artifacts["artifact4"]["id"],
        description="Not retrieved.",
    ).get_json()
    version5_response = actions.register_model_version(
        client,
        model_id=registered_models["model1"]["id"],
        artifact_id=registered_artifacts["artifact4"]["id"],
        description="new version",
    ).get_json()

    return {
        "version1": version1_response,
        "version2": version2_response,
        "version3": version3_response,
        "version4": version4_response,
        "version5": version5_response,
    }


@pytest.fixture
def registered_plugins(
    client: FlaskClient, db: SQLAlchemy, auth_account: dict[str, Any]
) -> dict[str, Any]:
    plugin1_response = actions.register_plugin(
        client,
        name="plugin_one",
        description="The first plugin.",
        group_id=auth_account["groups"][0]["id"],
    ).get_json()
    plugin2_response = actions.register_plugin(
        client,
        name="plugin_two",
        description="The second plugin.",
        group_id=auth_account["groups"][0]["id"],
    ).get_json()
    plugin3_response = actions.register_plugin(
        client,
        name="plugin_three",
        description="Not retrieved.",
        group_id=auth_account["groups"][0]["id"],
    ).get_json()
    return {
        "plugin1": plugin1_response,
        "plugin2": plugin2_response,
        "plugin3": plugin3_response,
    }


@pytest.fixture
def registered_plugin_with_files(
    client: FlaskClient, db: SQLAlchemy, auth_account: dict[str, Any]
) -> dict[str, Any]:
    plugin_response = actions.register_plugin(
        client,
        name="plugin",
        description="The plugin with files.",
        group_id=auth_account["default_group_id"],
    ).get_json()
    contents = textwrap.dedent(
        """from dioptra import pyplugs

        @pyplugs.register
        def hello_world(name: str) -> str:
            return f"Hello, {name}!"
        """
    )

    plugin_file1_response = actions.register_plugin_file(
        client,
        plugin_id=plugin_response["id"],
        filename="plugin_file_one.py",
        description="The first plugin file.",
        contents=contents,
    ).get_json()
    plugin_file2_response = actions.register_plugin_file(
        client,
        plugin_id=plugin_response["id"],
        filename="plugin_file_two.py",
        description="The second plugin file.",
        contents=contents,
    ).get_json()
    plugin_file3_response = actions.register_plugin_file(
        client,
        plugin_id=plugin_response["id"],
        filename="plugin_file_three.py",
        description="Not Retrieved.",
        contents=contents,
    ).get_json()
    return {
        "plugin": plugin_response,
        "plugin_file1": plugin_file1_response,
        "plugin_file2": plugin_file2_response,
        "plugin_file3": plugin_file3_response,
    }


@pytest.fixture
def registered_plugin_with_file_and_tasks(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
) -> dict[str, Any]:
    plugin_response = actions.register_plugin(
        client,
        name="plugin",
        description="The plugin with files.",
        group_id=auth_account["default_group_id"],
    ).get_json()
    contents = textwrap.dedent(
        """from dioptra import pyplugs

        @pyplugs.register
        def hello_world(name: str) -> str:
            return f"Hello, {name}!"


        @pyplugs.register
        def add_one(x: int) -> int:
            return x + 1
        """
    )
    string_type_response = actions.register_plugin_parameter_type(
        client,
        name="string",
        group_id=auth_account["default_group_id"],
        structure=None,
        description="The plugin parameter string type.",
    ).get_json()
    integer_type_response = actions.register_plugin_parameter_type(
        client,
        name="integer",
        group_id=auth_account["default_group_id"],
        structure=None,
        description="The plugin parameter integer type.",
    ).get_json()
    hello_world_task = {
        "name": "hello_world",
        "inputParams": [{"name": "name", "parameterType": string_type_response["id"]}],
        "outputParams": [
            {
                "name": "hello_world_message",
                "parameterType": string_type_response["id"],
            }
        ],
    }
    add_one_task = {
        "name": "add_one",
        "inputParams": [{"name": "x", "parameterType": integer_type_response["id"]}],
        "outputParams": [
            {
                "name": "value",
                "parameterType": integer_type_response["id"],
            }
        ],
    }
    plugin_task_list = [hello_world_task, add_one_task]
    plugin_file_response = actions.register_plugin_file(
        client,
        plugin_id=plugin_response["id"],
        description="The plugin file with tasks.",
        filename="plugin_file.py",
        contents=contents,
        tasks=plugin_task_list,
    ).get_json()
    return {
        "plugin": plugin_response,
        "plugin_file": plugin_file_response,
        "string_parameter_type": string_type_response,
        "integer_parameter_type": integer_type_response,
        "tasks": plugin_task_list,
    }


@pytest.fixture
def registered_queues(
    client: FlaskClient, db: SQLAlchemy, auth_account: dict[str, Any]
) -> dict[str, Any]:
    queue1_response = actions.register_queue(
        client,
        name="tensorflow_cpu",
        description="The first queue.",
        group_id=auth_account["groups"][0]["id"],
    ).get_json()
    queue2_response = actions.register_queue(
        client,
        name="tensorflow_gpu",
        description="The second queue.",
        group_id=auth_account["groups"][0]["id"],
    ).get_json()
    queue3_response = actions.register_queue(
        client,
        name="pytorch_cpu",
        description="Not retrieved.",
        group_id=auth_account["groups"][0]["id"],
    ).get_json()
    return {
        "queue1": queue1_response,
        "queue2": queue2_response,
        "queue3": queue3_response,
    }


@pytest.fixture
def registered_groups(
    client: FlaskClient, db: SQLAlchemy, auth_account: dict[str, Any]
) -> dict[str, Any]:
    public_response = actions.get_public_group(client).get_json()
    # group1_response = actions.register_group(
    # client,
    #  name="group_one",
    # ).get_json()
    # group2_response = actions.register_group(
    # client,
    # name="group_two",
    # ).get_json()
    # group3_response = actions.register_group(
    # client,
    # name="group_three",
    # ).get_json()
    return {
        "public": public_response,
        # "group1": group1_response,
        # "group2": group2_response,
        # "group3": group3_response,
    }


@pytest.fixture
def registered_plugin_parameter_types(
    client: FlaskClient, db: SQLAlchemy, auth_account: dict[str, Any]
) -> dict[str, Any]:
    built_in_types = {"any", "number", "integer", "string", "boolean", "null"}
    response = actions.get_plugin_parameter_types(client).get_json()
    built_in_types_dict = {
        param_type["name"]: param_type
        for param_type in response["data"]
        if param_type["name"] in built_in_types
    }
    plugin_param_type1_response = actions.register_plugin_parameter_type(
        client,
        name="image_shape",
        group_id=auth_account["groups"][0]["id"],
        structure={"list": ["integer"]},
        description="The dimensions of an image",
    ).get_json()
    plugin_param_type2_response = actions.register_plugin_parameter_type(
        client,
        name="model_output",
        group_id=auth_account["groups"][0]["id"],
        structure=dict({"list": ["float"]}),
        description="The softmax scores from a model",
    ).get_json()
    plugin_param_type3_response = actions.register_plugin_parameter_type(
        client,
        name="model",
        group_id=auth_account["groups"][0]["id"],
        structure=dict(),
        description="Opaque type for an ml model",
    ).get_json()
    return {
        **built_in_types_dict,
        "plugin_param_type1": plugin_param_type1_response,
        "plugin_param_type2": plugin_param_type2_response,
        "plugin_param_type3": plugin_param_type3_response,
    }


@pytest.fixture
def registered_experiments(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_entrypoints: dict[str, Any],
) -> dict[str, Any]:
    entrypoint_ids = [
        entrypoint["id"] for entrypoint in registered_entrypoints.values()
    ]
    experiment1_response = actions.register_experiment(
        client,
        name="experiment1",
        group_id=auth_account["default_group_id"],
        description="test description",
        entrypoint_ids=entrypoint_ids,
    ).get_json()
    experiment2_response = actions.register_experiment(
        client,
        name="experiment2",
        group_id=auth_account["default_group_id"],
        description="test description",
        entrypoint_ids=entrypoint_ids,
    ).get_json()
    experiment3_response = actions.register_experiment(
        client,
        name="experiment3",
        group_id=auth_account["default_group_id"],
        description="test description",
        entrypoint_ids=entrypoint_ids,
    ).get_json()
    return {
        "experiment1": experiment1_response,
        "experiment2": experiment2_response,
        "experiment3": experiment3_response,
    }


@pytest.fixture
def registered_entrypoints(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_queues: dict[str, Any],
    registered_plugin_with_files: dict[str, Any],
) -> dict[str, Any]:
    task_graph = textwrap.dedent(
        """# my entrypoint graph
        graph:
          message:
            my_entrypoint: $name
        """
    )
    parameters = [
        {
            "name": "entrypoint_param_1",
            "defaultValue": "default",
            "parameterType": "string",
        },
        {
            "name": "entrypoint_param_2",
            "defaultValue": "1.0",
            "parameterType": "float",
        },
        {
            "name": "entrypoint_param_3",
            "defaultValue": "/path",
            "parameterType": "string",
        },
        {
            "name": "entrypoint_param_4",
            "defaultValue": "1",
            "parameterType": "integer",
        },
        {
            "name": "entrypoint_param_5",
            "defaultValue": "['a', 'b', {'c': 1}]",
            "parameterType": "list",
        },
    ]
    plugin_ids = [registered_plugin_with_files["plugin"]["id"]]
    queue_ids = [queue["id"] for queue in list(registered_queues.values())]
    entrypoint1_response = actions.register_entrypoint(
        client,
        name="entrypoint_one",
        description="The first entrypoint.",
        group_id=auth_account["groups"][0]["id"],
        task_graph=task_graph,
        parameters=parameters,
        plugin_ids=plugin_ids,
        queue_ids=queue_ids,
    ).get_json()
    entrypoint2_response = actions.register_entrypoint(
        client,
        name="entrypoint_two",
        description="The second entrypoint.",
        group_id=auth_account["groups"][0]["id"],
        task_graph=task_graph,
        parameters=parameters,
        plugin_ids=plugin_ids,
        queue_ids=queue_ids,
    ).get_json()
    entrypoint3_response = actions.register_entrypoint(
        client,
        name="entrypoint_three",
        description="Not retrieved.",
        group_id=auth_account["groups"][0]["id"],
        task_graph=task_graph,
        parameters=parameters,
        plugin_ids=plugin_ids,
        queue_ids=queue_ids,
    ).get_json()
    return {
        "entrypoint1": entrypoint1_response,
        "entrypoint2": entrypoint2_response,
        "entrypoint3": entrypoint3_response,
    }


@pytest.fixture
def registered_jobs(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_queues: dict[str, Any],
    registered_experiments: dict[str, Any],
    registered_entrypoints: dict[str, Any],
    monkeypatch: MonkeyPatch,
) -> dict[str, Any]:
    # Inline import necessary to prevent circular import
    import dioptra.restapi.v1.shared.rq_service as rq_service

    monkeypatch.setattr(rq_service, "RQQueue", mock_rq.MockRQQueue)

    queue_id = registered_queues["queue1"]["snapshot"]
    experiment_id = registered_experiments["experiment1"]["snapshot"]
    entrypoint_id = registered_entrypoints["entrypoint1"]["snapshot"]
    values = {
        registered_entrypoints["entrypoint1"]["parameters"][0]["name"]: "new_value",
    }
    timeout = "24h"
    job1_response = actions.register_job(
        client=client,
        queue_id=queue_id,
        experiment_id=experiment_id,
        entrypoint_id=entrypoint_id,
        description="The first job.",
        values=values,
        timeout=timeout,
    ).get_json()
    job2_response = actions.register_job(
        client=client,
        queue_id=queue_id,
        experiment_id=experiment_id,
        entrypoint_id=entrypoint_id,
        description="The second job.",
        values=values,
        timeout=timeout,
    ).get_json()
    job3_response = actions.register_job(
        client=client,
        queue_id=queue_id,
        experiment_id=experiment_id,
        entrypoint_id=entrypoint_id,
        description="Not retrieved.",
        values=values,
        timeout=timeout,
    ).get_json()
    return {
        "job1": job1_response,
        "job2": job2_response,
        "job3": job3_response,
    }



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""Test suite for tag operations.

This module contains a set of tests that validate the CRUD operations and additional
functionalities for the tag entity. The tests ensure that the tags can be
registered, renamed, deleted, and queried as expected through the REST API.
"""
from typing import Any

import pytest
from flask.testing import FlaskClient
from flask_sqlalchemy import SQLAlchemy
from werkzeug.test import TestResponse

from dioptra.restapi.routes import V1_ROOT, V1_TAGS_ROUTE

from ..lib import actions, helpers

# -- Actions ---------------------------------------------------------------------------


def modify_tag(
    client: FlaskClient,
    tag_id: int,
    new_name: str,
) -> TestResponse:
    """Rename a tag using the API.

    Args:
        client: The Flask test client.
        tag_id: The id of the tag to rename.
        new_name: The new name to assign to the tag.

    Returns:
        The response from the API.
    """
    payload: dict[str, Any] = {"name": new_name}

    return client.put(
        f"/{V1_ROOT}/{V1_TAGS_ROUTE}/{tag_id}",
        json=payload,
        follow_redirects=True,
    )


def delete_tag(
    client: FlaskClient,
    tag_id: int,
) -> TestResponse:
    """Delete a tag using the API.

    Args:
        client: The Flask test client.
        tag_id: The id of the tag to delete.

    Returns:
        The response from the API.
    """

    return client.delete(
        f"/{V1_ROOT}/{V1_TAGS_ROUTE}/{tag_id}",
        follow_redirects=True,
    )


# -- Assertions ------------------------------------------------------------------------


def assert_tag_response_contents_matches_expectations(
    response: dict[str, Any], expected_contents: dict[str, Any]
) -> None:
    expected_keys = {
        "id",
        "group",
        "user",
        "createdOn",
        "lastModifiedOn",
        "name",
    }
    assert set(response.keys()) == expected_keys

    # Validate the non-Ref fields
    assert isinstance(response["id"], int)
    assert isinstance(response["name"], str)
    assert isinstance(response["createdOn"], str)
    assert isinstance(response["lastModifiedOn"], str)

    assert response["name"] == expected_contents["name"]

    assert helpers.is_iso_format(response["createdOn"])
    assert helpers.is_iso_format(response["lastModifiedOn"])

    # Validate the UserRef structure
    assert isinstance(response["user"]["id"], int)
    assert isinstance(response["user"]["username"], str)
    assert isinstance(response["user"]["url"], str)
    assert response["user"]["id"] == expected_contents["user_id"]

    # Validate the GroupRef structure
    assert isinstance(response["group"]["id"], int)
    assert isinstance(response["group"]["name"], str)
    assert isinstance(response["group"]["url"], str)
    assert response["group"]["id"] == expected_contents["group_id"]


def assert_retrieving_tag_by_id_works(
    client: FlaskClient,
    tag_id: int,
    expected: dict[str, Any],
) -> None:
    """Assert that retrieving a tag by id works.

    Args:
        client: The Flask test client.
        tag_id: The id of the tag to retrieve.
        expected: The expected response from the API.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """
    response = client.get(f"/{V1_ROOT}/{V1_TAGS_ROUTE}/{tag_id}", follow_redirects=True)
    assert response.status_code == 200 and response.get_json() == expected


def assert_retrieving_tags_works(
    client: FlaskClient,
    expected: list[dict[str, Any]],
    group_id: int | None = None,
    search: str | None = None,
    paging_info: dict[str, Any] | None = None,
) -> None:
    """Assert that retrieving all tags works.

    Args:
        client: The Flask test client.
        expected: The expected response from the API.
        group_id: The group ID used in query parameters.
        search: The search string used in query parameters.
        paging_info: The paging information used in query parameters.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """

    query_string: dict[str, Any] = {}

    if group_id is not None:
        query_string["groupId"] = group_id

    if search is not None:
        query_string["search"] = search

    if paging_info is not None:
        query_string["index"] = paging_info["index"]
        query_string["pageLength"] = paging_info["page_length"]

    response = client.get(
        f"/{V1_ROOT}/{V1_TAGS_ROUTE}",
        query_string=query_string,
        follow_redirects=True,
    )
    assert response.status_code == 200 and response.get_json()["data"] == expected


def assert_sorting_tag_works(
    client: FlaskClient,
    sortBy: str,
    descending: bool,
    expected: list[str],
) -> None:
    """Assert that tags can be sorted by column ascending/descending.

    Args:
        client: The Flask test client.
        expected: The expected order of tag ids after sorting.
            See test_tag_sort for expected orders.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """

    query_string: dict[str, Any] = {}

    query_string["sortBy"] = sortBy
    query_string["descending"] = descending

    response = client.get(
        f"/{V1_ROOT}/{V1_TAGS_ROUTE}",
        query_string=query_string,
        follow_redirects=True,
    )

    response_data = response.get_json()
    tag_ids = [tag["id"] for tag in response_data["data"]]

    assert response.status_code == 200 and tag_ids == expected


def assert_registering_existing_tag_name_fails(
    client: FlaskClient, name: str, group_id: int
) -> None:
    """Assert that registering a tag with an existing name fails.

    Args:
        client: The Flask test client.
        name: The name to assign to the new tag.

    Raises:
        AssertionError: If the response status code is not 400.
    """
    response = actions.register_tag(client, name=name, group_id=group_id)
    assert response.status_code == 409


def assert_tag_name_matches_expected_name(
    client: FlaskClient, tag_id: int, expected_name: str
) -> None:
    """Assert that the name of a tag matches the expected name.

    Args:
        client: The Flask test client.
        tag_id: The id of the tag to retrieve.
        expected_name: The expected name of the tag.

    Raises:
        AssertionError: If the response status code is not 200 or if the name of the
            tag does not match the expected name.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_TAGS_ROUTE}/{tag_id}",
        follow_redirects=True,
    )
    assert response.status_code == 200 and response.get_json()["name"] == expected_name


def assert_tag_is_not_found(
    client: FlaskClient,
    tag_id: int,
) -> None:
    """Assert that a tag is not found.

    Args:
        client: The Flask test client.
        tag_id: The id of the tag to retrieve.

    Raises:
        AssertionError: If the response status code is not 404.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_TAGS_ROUTE}/{tag_id}",
        follow_redirects=True,
    )
    assert response.status_code == 404


def assert_cannot_rename_tag_with_existing_name(
    client: FlaskClient,
    tag_id: int,
    existing_name: str,
) -> None:
    """Assert that renaming a tag with an existing name fails.

    Args:
        client: The Flask test client.
        tag_id: The id of the tag to rename.
        name: The name of an existing tag.

    Raises:
        AssertionError: If the response status code is not 400.
    """
    response = modify_tag(
        client=client,
        tag_id=tag_id,
        new_name=existing_name,
    )
    assert response.status_code == 409


# -- Tests -----------------------------------------------------------------------------


def test_create_tag(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
) -> None:
    """Test that tags can be registered and retrieved using the API.

    This test validates the following sequence of actions:

    - A user registers a tag: {name: "tag"}
    - The user is able to retrieve information about the tag using the
      tag id.
    - The returned information matches the information that was provided
      during registration.
    """
    name = "tag"
    user_id = auth_account["default_group_id"]
    group_id = auth_account["default_group_id"]
    tag_response = actions.register_tag(
        client,
        name=name,
        group_id=group_id,
    )
    tag_expected = tag_response.get_json()

    assert_tag_response_contents_matches_expectations(
        response=tag_expected,
        expected_contents={
            "name": name,
            "user_id": user_id,
            "group_id": group_id,
        },
    )
    assert_retrieving_tag_by_id_works(
        client, tag_id=tag_expected["id"], expected=tag_expected
    )


@pytest.mark.parametrize(
    "sortBy, descending , expected",
    [
        (None, None, ["tag1", "tag2", "tag3"]),
        ("name", True, ["tag2", "tag1", "tag3"]),
        ("name", False, ["tag3", "tag1", "tag2"]),
        ("createdOn", True, ["tag3", "tag2", "tag1"]),
        ("createdOn", False, ["tag1", "tag2", "tag3"]),
    ],
)
def test_tag_sort(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_tags: dict[str, Any],
    sortBy: str,
    descending: bool,
    expected: list[str],
) -> None:
    """Test that tags can be sorted by column.

    Given an authenticated user and registered tags, this test validates the following
    sequence of actions:

    - A user registers three tags, "tag_one", "tag_two", "name".
    - The user is able to retrieve a list of all registered tags sorted by a column
      ascending/descending.
    - The returned list of tags matches the order in the parametrize lists above.
    """

    expected_ids = [registered_tags[expected_name]["id"] for expected_name in expected]
    assert_sorting_tag_works(client, sortBy, descending, expected=expected_ids)


def test_tag_search_query(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_tags: dict[str, Any],
) -> None:
    """Test that tags can be registered and retrieved using the API with search terms.

    This test validates the following sequence of actions:

    - A user registers three tags: "tag1", "tag2", "tag3".
    - The user is able to retrieve a list of all registered tags with the string "tag"
      contained within the name.
    - In all cases, the returned information matches the information that was provided
      during registration.
    """
    tag1_expected = registered_tags["tag1"]
    tag2_expected = registered_tags["tag2"]
    tag_expected_list = [tag1_expected, tag2_expected]

    assert_retrieving_tags_works(
        client,
        expected=tag_expected_list,
        search="name:*tag*",
    )


def test_cannot_register_existing_tag_name(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_tags: dict[str, Any],
) -> None:
    """Test that registering a tag with an existing name fails.

    This test validates the following sequence of actions:

    - A user registers a tag named "tag1".
    - The user attempts to register a second tag with the same name, which fails.
    """
    existing_tag = registered_tags["tag1"]

    assert_registering_existing_tag_name_fails(
        client, name=existing_tag["name"], group_id=existing_tag["group"]["id"]
    )


def test_rename_tag(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_tags: dict[str, Any],
) -> None:
    """Test that a tag can be renamed.

    This test validates the following sequence of actions:

    - A user registers a tag named "ML: Object Recognition".
    - The user is able to retrieve information about the "ML: Object Recognition" tag
      that matches the information that was provided during registration.
    - The user renames this same tag to "new_name".
    - The user retrieves information about the same tag and it reflects the name
      change.
    """
    updated_tag_name = "new_name"
    tag_to_rename = registered_tags["tag1"]
    existing_tag = registered_tags["tag2"]

    modify_tag(
        client,
        tag_id=tag_to_rename["id"],
        new_name=updated_tag_name,
    )
    assert_tag_name_matches_expected_name(
        client, tag_id=tag_to_rename["id"], expected_name=updated_tag_name
    )
    assert_cannot_rename_tag_with_existing_name(
        client,
        tag_id=tag_to_rename["id"],
        existing_name=existing_tag["name"],
    )


def test_delete_tag_by_id(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_tags: dict[str, Any],
) -> None:
    """Test that a tag can be deleted by referencing its id.

    This test validates the following sequence of actions:

    - A user registers a tag.
    - The user is able to retrieve information about the tag that
      matches the information that was provided during registration.
    - The user deletes the tag by referencing its id.
    - The user attempts to retrieve information about the tag, which
      is no longer found.
    """
    tag_expected = registered_tags["tag1"]

    assert_retrieving_tag_by_id_works(
        client, tag_id=tag_expected["id"], expected=tag_expected
    )
    delete_tag(client, tag_id=tag_expected["id"])
    assert_tag_is_not_found(client, tag_id=tag_expected["id"])



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The module defining the endpoints for Experiment resources."""
import uuid
from typing import cast
from urllib.parse import unquote

import structlog
from flask import request
from flask_accepts import accepts, responds
from flask_login import login_required
from flask_restx import Namespace, Resource
from injector import inject
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import models
from dioptra.restapi.routes import V1_EXPERIMENTS_ROUTE
from dioptra.restapi.v1 import utils
from dioptra.restapi.v1.artifacts.schema import ArtifactSchema
from dioptra.restapi.v1.artifacts.service import JobArtifactService
from dioptra.restapi.v1.entrypoints.schema import EntrypointRefSchema
from dioptra.restapi.v1.jobs.schema import (
    ExperimentJobGetQueryParameters,
    JobMlflowRunSchema,
    JobPageSchema,
    JobSchema,
    JobStatusSchema,
)
from dioptra.restapi.v1.jobs.service import (
    ExperimentJobIdMlflowrunService,
    ExperimentJobIdService,
    ExperimentJobIdStatusService,
    ExperimentJobService,
)
from dioptra.restapi.v1.schemas import IdListSchema, IdStatusResponseSchema
from dioptra.restapi.v1.shared.drafts.controller import (
    generate_resource_drafts_endpoint,
    generate_resource_drafts_id_endpoint,
    generate_resource_id_draft_endpoint,
)
from dioptra.restapi.v1.shared.snapshots.controller import (
    generate_resource_snapshots_endpoint,
    generate_resource_snapshots_id_endpoint,
)
from dioptra.restapi.v1.shared.tags.controller import (
    generate_resource_tags_endpoint,
    generate_resource_tags_id_endpoint,
)

from .schema import (
    ExperimentDraftSchema,
    ExperimentGetQueryParameters,
    ExperimentMutableFieldsSchema,
    ExperimentPageSchema,
    ExperimentSchema,
)
from .service import (
    RESOURCE_TYPE,
    SEARCHABLE_FIELDS,
    ExperimentIdEntrypointsIdService,
    ExperimentIdEntrypointsService,
    ExperimentIdService,
    ExperimentService,
)

LOGGER: BoundLogger = structlog.stdlib.get_logger()

api: Namespace = Namespace("Experiments", description="Experiments endpoint")


@api.route("/")
class ExperimentEndpoint(Resource):
    @inject
    def __init__(self, experiment_service: ExperimentService, *args, **kwargs) -> None:
        """Initialize the experiment resource.

        All arguments are provided via dependency injection.

        Args:
            experiment_service: An ExperimentService object.
        """
        self._experiment_service = experiment_service
        super().__init__(*args, **kwargs)

    @login_required
    @accepts(query_params_schema=ExperimentGetQueryParameters, api=api)
    @responds(schema=ExperimentPageSchema, api=api)
    def get(self):
        """Gets a page of Experiment resources matching the filter parameters."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="ExperimentEndpoint",
            request_type="GET",
        )
        parsed_query_params = request.parsed_query_params  # noqa: F841

        group_id = parsed_query_params["group_id"]
        search_string = parsed_query_params["search"]
        page_index = parsed_query_params["index"]
        page_length = parsed_query_params["page_length"]
        sort_by_string = parsed_query_params["sort_by"]
        descending = parsed_query_params["descending"]

        experiments, total_num_experiments = self._experiment_service.get(
            group_id=group_id,
            search_string=search_string,
            page_index=page_index,
            page_length=page_length,
            sort_by_string=sort_by_string,
            descending=descending,
            log=log,
        )
        return utils.build_paging_envelope(
            "experiments",
            build_fn=utils.build_experiment,
            data=experiments,
            group_id=group_id,
            query=search_string,
            draft_type=None,
            index=page_index,
            length=page_length,
            total_num_elements=total_num_experiments,
            sort_by=sort_by_string,
            descending=descending,
        )

    @login_required
    @accepts(schema=ExperimentSchema, api=api)
    @responds(schema=ExperimentSchema, api=api)
    def post(self):
        """Creates an Experiment resource with a provided name."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="ExperimentEndpoint",
            request_type="POST",
        )
        parsed_obj = request.parsed_obj  # noqa: F841
        experiment = self._experiment_service.create(
            name=parsed_obj["name"],
            description=parsed_obj["description"],
            group_id=parsed_obj["group_id"],
            entrypoint_ids=parsed_obj["entrypoint_ids"],
            log=log,
        )
        return utils.build_experiment(experiment)


@api.route("/<int:id>")
@api.param("id", "ID for the Experiment resource.")
class ExperimentIdEndpoint(Resource):
    @inject
    def __init__(
        self, experiment_id_service: ExperimentIdService, *args, **kwargs
    ) -> None:
        """Initialize the experiment resource.

        All arguments are provided via dependency injection.

        Args:
            experiment_id_service: An ExperimentIdService object.
        """
        self._experiment_id_service = experiment_id_service
        super().__init__(*args, **kwargs)

    @login_required
    @responds(schema=ExperimentSchema, api=api)
    def get(self, id: int):
        """Gets an experiment by its unique identifier."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="Experiment",
            request_type="GET",
            id=id,
        )
        experiment = cast(
            models.Experiment,
            self._experiment_id_service.get(id, error_if_not_found=True, log=log),
        )
        return utils.build_experiment(experiment)

    @login_required
    @responds(schema=IdStatusResponseSchema, api=api)
    def delete(self, id: int):
        """Deletes an experiment by its unique identifier."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="Experiment",
            request_type="DELETE",
            id=id,
        )
        return self._experiment_id_service.delete(id, log=log)

    @login_required
    @accepts(schema=ExperimentMutableFieldsSchema, api=api)
    @responds(schema=ExperimentSchema, api=api)
    def put(self, id: int):
        """Modifies an experiment by its unique identifier."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="Experiment",
            request_type="PUT",
            id=id,
        )
        parsed_obj = request.parsed_obj  # type: ignore
        experiment = cast(
            utils.ExperimentDict,
            self._experiment_id_service.modify(
                id,
                name=parsed_obj["name"],
                description=parsed_obj["description"],
                entrypoint_ids=parsed_obj["entrypoint_ids"],
                error_if_not_found=True,
                log=log,
            ),
        )
        return utils.build_experiment(experiment)


@api.route("/<int:id>/jobs")
@api.param("id", "ID for the Experiment resource.")
class ExperimentIdJobEndpoint(Resource):
    @inject
    def __init__(
        self,
        experiment_job_service: ExperimentJobService,
        *args,
        **kwargs,
    ) -> None:
        """Initialize the Job resource.

        All arguments are provided via dependency injection.

        Args:
            experiment_id_job_service: An ExperimentIdJobService object.
        """
        self._experiment_job_service = experiment_job_service
        super().__init__(*args, **kwargs)

    @login_required
    @accepts(query_params_schema=ExperimentJobGetQueryParameters, api=api)
    @responds(schema=JobPageSchema, api=api)
    def get(self, id: int):
        """Returns a list of jobs for a specified Experiment."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="ExperimentIdJobEndpoint",
            request_type="GET",
            id=id,
        )
        parsed_query_params = request.parsed_query_params  # type: ignore

        search_string = unquote(parsed_query_params["search"])
        page_index = parsed_query_params["index"]
        page_length = parsed_query_params["page_length"]
        sort_by_string = unquote(parsed_query_params["sort_by"])
        descending = parsed_query_params["descending"]

        jobs, total_num_jobs = self._experiment_job_service.get(
            experiment_id=id,
            search_string=search_string,
            page_index=page_index,
            page_length=page_length,
            sort_by_string=sort_by_string,
            descending=descending,
            log=log,
        )
        return utils.build_paging_envelope(
            f"experiments/{id}/jobs",
            build_fn=utils.build_job,
            data=jobs,
            group_id=None,
            query=search_string,
            draft_type=None,
            index=page_index,
            length=page_length,
            total_num_elements=total_num_jobs,
            sort_by=sort_by_string,
            descending=descending,
        )

    @login_required
    @accepts(schema=JobSchema(exclude=["groupId"]), api=api)
    @responds(schema=JobSchema, api=api)
    def post(self, id: int):
        """Creates a Job resource under the specified Experiment."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="ExperimentIdJobEndpoint",
            request_type="POST",
        )
        parsed_obj = request.parsed_obj  # type: ignore

        job = self._experiment_job_service.create(
            experiment_id=id,
            queue_id=parsed_obj["queue_id"],
            entrypoint_id=parsed_obj["entrypoint_id"],
            values=parsed_obj["values"],
            description=parsed_obj["description"],
            timeout=parsed_obj["timeout"],
            log=log,
        )
        return utils.build_job(job)


@api.route("/<int:id>/jobs/<int:jobId>")
@api.param("id", "ID for the Experiment resource.")
@api.param("jobId", "ID for the Job resource.")
class ExperimentIdJobIdEndpoint(Resource):
    @inject
    def __init__(
        self, experiment_job_id_service: ExperimentJobIdService, *args, **kwargs
    ) -> None:
        """Initialize the jobs resource.

        All arguments are provided via dependency injection.

        Args:
            job_id_service: A JobIdService object.
        """
        self._experiment_job_id_service = experiment_job_id_service
        super().__init__(*args, **kwargs)

    @login_required
    @responds(schema=JobSchema, api=api)
    def get(self, id: int, jobId: int):
        """Gets a Job resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="ExperimentIdJobIdEndpoint",
            request_type="GET",
            id=id,
            job_id=jobId,
        )
        job = self._experiment_job_id_service.get(id, job_id=jobId, log=log)
        return utils.build_job(job)

    @login_required
    @responds(schema=IdStatusResponseSchema, api=api)
    def delete(self, id: int, jobId: int):
        """Deletes a Job resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="ExperimentIdJobIdEndpoint",
            request_type="DELETE",
            id=id,
            job_id=jobId,
        )
        return self._experiment_job_id_service.delete(id, job_id=jobId, log=log)


@api.route("/<int:id>/jobs/<int:jobId>/status")
@api.param("id", "ID for the Experiment resource.")
@api.param("jobId", "ID for the Job resource.")
class ExperimentIdJobIdStatusEndpoint(Resource):
    @inject
    def __init__(
        self,
        experiment_job_id_status_service: ExperimentJobIdStatusService,
        *args,
        **kwargs,
    ) -> None:
        """Initialize the jobs resource.

        All arguments are provided via dependency injection.

        Args:
            job_id_service: A JobIdStatusService object.
        """
        self._experiment_job_id_status_service = experiment_job_id_status_service
        super().__init__(*args, **kwargs)

    @login_required
    @responds(schema=JobStatusSchema, api=api)
    def get(self, id: int, jobId: int):
        """Gets a Job resource's status."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="ExperimentIdJobIdStatusEndpoint",
            request_type="GET",
            experiment_id=id,
            job_id=jobId,
        )
        return self._experiment_job_id_status_service.get(
            experiment_id=id, job_id=jobId, error_if_not_found=True, log=log
        )

    @login_required
    @accepts(schema=JobStatusSchema, api=api)
    @responds(schema=JobStatusSchema, api=api)
    def put(self, id: int, jobId: int):
        """Modifies a Job resource's status."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="ExperimentIdJobIdStatusEndpoint",
            request_type="PUT",
            experiment_id=id,
            job_id=jobId,
        )
        parsed_obj = request.parsed_obj  # type: ignore
        return self._experiment_job_id_status_service.modify(
            experiment_id=id,
            job_id=jobId,
            status=parsed_obj["status"],
            error_if_not_found=True,
            log=log,
        )


@api.route("/<int:id>/jobs/<int:jobId>/mlflowRun")
@api.param("id", "ID for the Experiment resource.")
@api.param("jobId", "ID for the Job resource.")
class ExperimentIdJobIdMlflowrunEndpoint(Resource):
    @inject
    def __init__(
        self,
        experiment_job_id_mlflowrun_service: ExperimentJobIdMlflowrunService,
        *args,
        **kwargs,
    ) -> None:
        """Initialize the jobs resource.

        All arguments are provided via dependency injection.

        Args:
            job_id_service: A JobIdStatusService object.
        """
        self._experiment_job_id_mlflowrun_service = experiment_job_id_mlflowrun_service
        super().__init__(*args, **kwargs)

    @login_required
    @responds(schema=JobMlflowRunSchema, api=api)
    def get(self, id: int, jobId: int):
        """Gets a Job resource's mlflow run id."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="ExperimentIdJobIdMlflowrunEndpoint",
            request_type="GET",
            experiment_id=id,
            job_id=jobId,
        )
        return self._experiment_job_id_mlflowrun_service.get(
            experiment_id=id, job_id=jobId, error_if_not_found=True, log=log
        )

    @login_required
    @accepts(schema=JobMlflowRunSchema, api=api)
    @responds(schema=JobMlflowRunSchema, api=api)
    def post(self, id: int, jobId: int):
        """Sets the mlflow run id for a Job"""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="ExperimentIdJobIdMlflowrunEndpoint",
            request_type="POST",
            experiment_id=id,
            job_id=jobId,
        )
        parsed_obj = request.parsed_obj  # type: ignore
        return self._experiment_job_id_mlflowrun_service.create(
            experiment_id=id,
            job_id=jobId,
            mlflow_run_id=parsed_obj["mlflow_run_id"],
            error_if_not_found=True,
            log=log,
        )


@api.route("/<int:id>/jobs/<int:jobId>/artifacts")
@api.param("id", "ID for the Experiment resource.")
@api.param("jobId", "ID for the Job resource.")
class ExperimentIdJobIdArtifactsEndpoint(Resource):
    @inject
    def __init__(
        self, job_artifact_service: JobArtifactService, *args, **kwargs
    ) -> None:
        """Initialize the jobs resource.

        All arguments are provided via dependency injection.

        Args:
            job_id_service: A JobIdStatusService object.
        """
        self._job_artifact_service = job_artifact_service
        super().__init__(*args, **kwargs)

    @login_required
    @accepts(
        schema=ArtifactSchema(exclude=["groupId", "jobId"]),
        model_name="JobArtifactSchema",
        api=api,
    )
    @responds(schema=ArtifactSchema, api=api)
    def post(self, id: int, jobId: int):
        """Creates an Artifact resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Artifact", request_type="POST"
        )
        log.debug("Request received")
        parsed_obj = request.parsed_obj  # type: ignore

        artifact = self._job_artifact_service.create(
            experiment_id=id,
            job_id=jobId,
            uri=parsed_obj["uri"],
            description=parsed_obj["description"],
            log=log,
        )
        return utils.build_artifact(artifact)


@api.route("/<int:id>/entrypoints")
@api.param("id", "ID for the Experiment resource.")
class ExperimentIdEntrypointsEndpoint(Resource):
    @inject
    def __init__(
        self,
        experiment_id_entrypoints_service: ExperimentIdEntrypointsService,
        *args,
        **kwargs,
    ) -> None:
        """Initialize the experiment resource.

        All arguments are provided via dependency injection.

        Args:
            experiment_id_entrypoints_service: An ExperimentIdEntrypointsService object.
        """
        self._experiment_id_entrypoints = experiment_id_entrypoints_service
        super().__init__(*args, **kwargs)

    @login_required
    @responds(schema=EntrypointRefSchema(many=True), api=api)
    def get(self, id: int):
        """Gets the list of Entrypoints for the resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Experiment", request_type="GET"
        )
        entrypoints = self._experiment_id_entrypoints.get(id, log=log)
        return [utils.build_entrypoint_ref(entrypoint) for entrypoint in entrypoints]

    @login_required
    @accepts(schema=IdListSchema, api=api)
    @responds(schema=EntrypointRefSchema(many=True), api=api)
    def post(self, id: int):
        """Appends one or more Entrypoints to the resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Experiment", request_type="POST"
        )
        parsed_obj = request.parsed_obj  # type: ignore
        entrypoints = cast(
            list[models.EntryPoint],
            self._experiment_id_entrypoints.append(
                id, entrypoint_ids=parsed_obj["ids"], error_if_not_found=True, log=log
            ),
        )
        return [utils.build_entrypoint_ref(entrypoint) for entrypoint in entrypoints]

    @login_required
    @accepts(schema=IdListSchema, api=api)
    @responds(schema=EntrypointRefSchema(many=True), api=api)
    def put(self, id: int):
        """Replaces one or more Entrypoints to the resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Experiment", request_type="POST"
        )
        parsed_obj = request.parsed_obj  # type: ignore
        entrypoints = self._experiment_id_entrypoints.modify(
            id, entrypoint_ids=parsed_obj["ids"], error_if_not_found=True, log=log
        )
        return [utils.build_entrypoint_ref(entrypoint) for entrypoint in entrypoints]

    @login_required
    @responds(schema=IdStatusResponseSchema, api=api)
    def delete(self, id: int):
        """Removes all Entrypoints from the resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Experiment", request_type="DELETE"
        )
        return self._experiment_id_entrypoints.delete(
            id, error_if_not_found=True, log=log
        )


@api.route("/<int:id>/entrypoints/<int:entrypointId>")
@api.param("id", "ID for the Experiment resource.")
@api.param("entrypointId", "ID for the Entrypoint resource.")
class ExperimentIdEntrypointsId(Resource):
    @inject
    def __init__(
        self,
        experiments_id_entrypoints_id_service: ExperimentIdEntrypointsIdService,
        *args,
        **kwargs,
    ) -> None:
        self._experiments_id_entrypoints_id_service = (
            experiments_id_entrypoints_id_service
        )
        """Initialize the experiment resource.

        All arguments are provided via dependency injection.

        Args:
            experiment_id_entrypoints_id_service: An ExperimentIdEntrypointsIdService
            object.
        """
        super().__init__(*args, **kwargs)

    @login_required
    @responds(schema=IdStatusResponseSchema, api=api)
    def delete(self, id: int, entrypointId):
        """Removes a Entrypoint from the Experiment resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Experiment", request_type="GET"
        )
        return self._experiments_id_entrypoints_id_service.delete(
            id, entrypointId, log=log
        )


ExperimentDraftResource = generate_resource_drafts_endpoint(
    api,
    resource_name=RESOURCE_TYPE,
    route_prefix=V1_EXPERIMENTS_ROUTE,
    request_schema=ExperimentDraftSchema,
)
ExperimentDraftIdResource = generate_resource_drafts_id_endpoint(
    api,
    resource_name=RESOURCE_TYPE,
    request_schema=ExperimentDraftSchema(exclude=["groupId"]),
)
ExperimentIdDraftResource = generate_resource_id_draft_endpoint(
    api,
    resource_name=RESOURCE_TYPE,
    request_schema=ExperimentDraftSchema(exclude=["groupId"]),
)

ExperimentSnapshotsResource = generate_resource_snapshots_endpoint(
    api=api,
    resource_model=models.Experiment,
    resource_name=RESOURCE_TYPE,
    route_prefix=V1_EXPERIMENTS_ROUTE,
    searchable_fields=SEARCHABLE_FIELDS,
    page_schema=ExperimentPageSchema,
    build_fn=utils.build_experiment,
)
ExperimentSnapshotsIdResource = generate_resource_snapshots_id_endpoint(
    api=api,
    resource_model=models.Experiment,
    resource_name=RESOURCE_TYPE,
    response_schema=ExperimentSchema,
    build_fn=utils.build_experiment,
)

ExperimentTagsResource = generate_resource_tags_endpoint(
    api=api,
    resource_name=RESOURCE_TYPE,
)
ExperimentTagsIdResource = generate_resource_tags_id_endpoint(
    api=api, resource_name=RESOURCE_TYPE
)



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""Test suite for model operations.

This module contains a set of tests that validate the CRUD operations and additional
functionalities for the model entity. The tests ensure that the models can be
registered, renamed, deleted, and locked/unlocked as expected through the REST API.
"""

from typing import Any

import pytest
from flask.testing import FlaskClient
from flask_sqlalchemy import SQLAlchemy

from dioptra.restapi.routes import V1_ARTIFACTS_ROUTE, V1_ROOT

from ..lib import actions, helpers

# -- Assertions ------------------------------------------------------------------------


def assert_artifact_response_contents_matches_expectations(
    response: dict[str, Any],
    expected_contents: dict[str, Any],
) -> None:
    """Assert that artifact response contents is valid.

    Args:
        response: The actual response from the API.
        expected_contents: The expected response from the API.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response or if the response contents is not
            valid.
    """
    expected_keys = {
        "id",
        "snapshot",
        "group",
        "user",
        "createdOn",
        "lastModifiedOn",
        "latestSnapshot",
        "hasDraft",
        "uri",
        "description",
        "tags",
    }
    assert set(response.keys()) == expected_keys

    # Validate the non-Ref fields
    assert isinstance(response["id"], int)
    assert isinstance(response["snapshot"], int)
    assert isinstance(response["uri"], str)
    assert isinstance(response["description"], str)
    assert isinstance(response["createdOn"], str)
    assert isinstance(response["lastModifiedOn"], str)
    assert isinstance(response["latestSnapshot"], bool)
    assert isinstance(response["hasDraft"], bool)

    assert response["uri"] == expected_contents["uri"]
    assert response["description"] == expected_contents["description"]

    assert helpers.is_iso_format(response["createdOn"])
    assert helpers.is_iso_format(response["lastModifiedOn"])

    # Validate the UserRef structure
    assert isinstance(response["user"]["id"], int)
    assert isinstance(response["user"]["username"], str)
    assert isinstance(response["user"]["url"], str)
    assert response["user"]["id"] == expected_contents["user_id"]

    # Validate the GroupRef structure
    assert isinstance(response["group"]["id"], int)
    assert isinstance(response["group"]["name"], str)
    assert isinstance(response["group"]["url"], str)
    assert response["group"]["id"] == expected_contents["group_id"]

    # Validate the TagRef structure
    for tag in response["tags"]:
        assert isinstance(tag["id"], int)
        assert isinstance(tag["name"], str)
        assert isinstance(tag["url"], str)


def assert_retrieving_artifact_by_id_works(
    client: FlaskClient, artifact_id: int, expected: dict[str, Any]
) -> None:
    """Assert that retrieving a artifact by id works.

    Args:
        client: The Flask test client.
        artifact_id: The id of the artifact to retrieve.
        expected: The expected response from the API.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_ARTIFACTS_ROUTE}/{artifact_id}", follow_redirects=True
    )
    assert response.status_code == 200 and response.get_json() == expected


def assert_retrieving_artifacts_works(
    client: FlaskClient,
    expected: list[dict[str, Any]],
    group_id: int | None = None,
    search: str | None = None,
    paging_info: dict[str, Any] | None = None,
) -> None:
    """Assert that retrieving all artifacts works.

    Args:
        client: The Flask test client.
        expected: The expected response from the API.
        group_id: The group ID used in query parameters.
        search: The search string used in query parameters.
        paging_info: The paging information used in query parameters.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """
    query_string: dict[str, Any] = {}

    if group_id is not None:
        query_string["groupId"] = group_id

    if search is not None:
        query_string["search"] = search

    if paging_info is not None:
        query_string["index"] = paging_info["index"]
        query_string["pageLength"] = paging_info["page_length"]

    response = client.get(
        f"/{V1_ROOT}/{V1_ARTIFACTS_ROUTE}",
        query_string=query_string,
        follow_redirects=True,
    )
    assert response.status_code == 200 and response.get_json()["data"] == expected


def assert_sorting_artifact_works(
    client: FlaskClient,
    sortBy: str,
    descending: bool,
    expected: list[str],
) -> None:
    """Assert that artifacts can be sorted by column ascending/descending.

    Args:
        client: The Flask test client.
        expected: The expected order of artifacts ids after sorting.
            See test_artifact_sort for expected orders.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """

    query_string: dict[str, Any] = {}

    query_string["sortBy"] = sortBy
    query_string["descending"] = descending

    response = client.get(
        f"/{V1_ROOT}/{V1_ARTIFACTS_ROUTE}",
        query_string=query_string,
        follow_redirects=True,
    )

    response_data = response.get_json()
    artifact_ids = [artifact["id"] for artifact in response_data["data"]]

    assert response.status_code == 200 and artifact_ids == expected


def assert_registering_existing_artifact_uri_fails(
    client: FlaskClient,
    uri: str,
    group_id: int,
    job_id: int,
) -> None:
    """Assert that registering an artifact with an existing uri fails.

    Args:
        client: The Flask test client.
        uri: The uri to assign to the new artifact.

    Raises:
        AssertionError: If the response status code is not 400.
    """
    response = actions.register_artifact(
        client,
        uri=uri,
        description="",
        group_id=group_id,
        job_id=job_id,
    )
    assert response.status_code == 409


# -- Tests -----------------------------------------------------------------------------


def test_create_artifact(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_jobs: dict[str, Any],
) -> None:
    """Test that artifacts can be correctly registered and retrieved using the API.

    Given an authenticated user, this test validates the following sequence of actions:

    - The user registers an artifact with uri "s3://bucket/model_v1.artifact".
    - The response is valid matches the expected values given the registration request.
    - The user is able to retrieve information about the artifact using the artifact id.
    """
    uri = "s3://bucket/model_v1.artifact"
    description = "The first artifact."
    job_id = registered_jobs["job1"]["id"]
    user_id = auth_account["id"]
    group_id = auth_account["groups"][0]["id"]
    artifact_response = actions.register_artifact(
        client,
        uri=uri,
        job_id=job_id,
        group_id=group_id,
        description=description,
    )

    artifact_expected = artifact_response.get_json()

    assert_artifact_response_contents_matches_expectations(
        response=artifact_expected,
        expected_contents={
            "uri": uri,
            "description": description,
            "user_id": user_id,
            "group_id": group_id,
        },
    )
    assert_retrieving_artifact_by_id_works(
        client, artifact_id=artifact_expected["id"], expected=artifact_expected
    )


def test_artifacts_get_all(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_artifacts: dict[str, Any],
) -> None:
    """Test that all artifacts can be retrieved.

    Given an authenticated user and registered artifacts, this test validates the
    following sequence of actions:

    - A user registers three artifacts with uris
        - "s3://bucket/model_v1.artifact"
        - "s3://bucket/cnn.artifact"
        - "s3://bucket/model.artifact"
        - "s3://bucket/model_v2.artifact"
    - The user is able to retrieve a list of all registered artifacts.
    - The returned list of artifacts matches the full list of registered artifacts.
    """
    artifacts_expected_list = list(registered_artifacts.values())
    assert_retrieving_artifacts_works(client, expected=artifacts_expected_list)


@pytest.mark.parametrize(
    "sortBy, descending , expected",
    [
        (None, None, ["artifact1", "artifact2", "artifact3", "artifact4"]),
        ("description", True, ["artifact2", "artifact1", "artifact4", "artifact3"]),
        ("description", False, ["artifact3", "artifact4", "artifact1", "artifact2"]),
        ("createdOn", True, ["artifact4", "artifact3", "artifact2", "artifact1"]),
        ("createdOn", False, ["artifact1", "artifact2", "artifact3", "artifact4"]),
    ],
)
def test_artifact_sort(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_artifacts: dict[str, Any],
    sortBy: str,
    descending: bool,
    expected: list[str],
) -> None:
    """Test that artifacts can be sorted by column.

    Given an authenticated user and registered artifacts, this test validates the
    following sequence of actions:

    - A user registers three artifacts with these descriptions:
        "Model artifact.",
        "Trained conv net model artifact.",
        "Another model",
        "Fine-tuned model.".
    - The user is able to retrieve a list of all registered artifacts sorted by a column
    - The returned list of artifacts matches the order in the parametrize lists above.
    """

    expected_ids = [
        registered_artifacts[expected_name]["id"] for expected_name in expected
    ]
    assert_sorting_artifact_works(client, sortBy, descending, expected=expected_ids)


def test_artifact_search_query(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_artifacts: dict[str, Any],
) -> None:
    """Test that artifacts can be queried with a search term.

    Given an authenticated user and registered artifacts, this test validates the
    following sequence of actions:

    - The user is able to retrieve a list of all registered artifacts with 'artifact' in
        their description.
    - The returned list of artifacts matches the expected matches from the query.
    """
    artifacts_expected_list = list(registered_artifacts.values())[:2]
    assert_retrieving_artifacts_works(
        client,
        expected=artifacts_expected_list,
        search="description:*artifact*",
    )


def test_artifact_group_query(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_artifacts: dict[str, Any],
) -> None:
    """Test that artifacts can retrieved using a group filter.

    Given an authenticated user and registered artifacts, this test validates the
    following sequence of actions:

    - The user is able to retrieve a list of all registered artifacts that are owned by
      the default group.
    - The returned list of artifacts matches the expected list owned by the default
      group.
    """
    artifacts_expected_list = list(registered_artifacts.values())
    assert_retrieving_artifacts_works(
        client,
        expected=artifacts_expected_list,
        group_id=auth_account["groups"][0]["id"],
    )


def test_cannot_register_existing_artifact_uri(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_artifacts: dict[str, Any],
) -> None:
    """Test that registering a artifact with an existing uri fails.

    Given an authenticated user and registered artifacts, this test validates the
    following sequence of actions:

    - The user attempts to register a second artifact with the same uri.
    - The request fails with an appropriate error message and response code.
    """
    existing_artifact = registered_artifacts["artifact1"]
    assert_registering_existing_artifact_uri_fails(
        client,
        uri=existing_artifact["uri"],
        group_id=existing_artifact["group"]["id"],
        job_id=0,  # TODO: fill in once job stuff is done.
    )
